import glob
import logging
import os

import fsspec
import geopandas as gpd
import intake as intk
import pandas as pd
import ujson
import xarray as xr
import zarr
from kerchunk.hdf import SingleHdf5ToZarr
import hydrodatasource.configs.config as conf


def spec_path(url_path: str, head="local", need_cache=False, is_dir=False):
    """Access the file system to get data of specific path, file or directory.

    Parameters
    ----------
    url_path : str
        absolute path of the file or directory
    head : str, optional
        minio or local, means the where data source is, by default "local"
    need_cache : bool, optional
        _description_, by default False
    is_dir : bool, optional
        file or directory, by default False, means file

    Returns
    -------
    _type_
        _description_

    Raises
    ------
    ValueError
        _description_
    ValueError
        _description_
    """
    if is_dir is False:
        if head == "local":
            ret_data = read_valid_data(url_path, need_cache=need_cache)
        elif head == "minio":
            url_path = f"s3://{url_path}"
            ret_data = read_valid_data(
                url_path, storage_option=conf.MINIO_PARAM, need_cache=need_cache
            )
        else:
            raise ValueError("head should be 'local' or 'minio'")
    else:
        ret_data = []
        if head == "local":
            url_path = os.path.join(conf.LOCAL_DATA_PATH, url_path)
            ret_data.extend(
                read_valid_data(file, need_cache=need_cache)
                for file in glob.glob(url_path + "/**", recursive=True)
                if not os.path.isdir(file)
            )
        elif head == "minio":
            url_path = "s3://" + url_path
            ret_data.extend(
                read_valid_data(
                    file,
                    storage_option=conf.MINIO_PARAM,
                    need_cache=need_cache,
                )
                for file in conf.FS.glob(url_path + "/**")
                if not conf.FS.isdir(file)
            )
        else:
            raise ValueError("head should be 'local' or 'minio'")
    return ret_data


def read_valid_data(obj: str, storage_option=None, need_cache=False, need_refer=False):
    """
    Read valid data from different file types.
    See https://intake.readthedocs.io/en/latest/plugin-directory.html
    pip install intake-xarray intake-geopandas

    Parameters:
    ------------
    obj (str)
        The file path or URL of the data, format is 's3://bucket_name/directory_name/file_name' in minio.
    storage_option (dict, optional)
        The storage options for accessing the data. Defaults to None, if you want to
        read from minio, storage_option should be minio login params.
        See hydrodatasource.configs.config.MINIO_PARAM to get reference
    need_cache (bool, optional)
        Whether to cache the data. Defaults to False.

    Returns:
    ------------
    object: The data object.
    """
    data_obj = None
    dot_in_obj = "." in obj
    cache_name = obj.lstrip("s3://").split("/")[-1]
    if not dot_in_obj:
        data_obj = pd.read_fwf(obj, storage_options=storage_option)
        if (need_cache is True) & (storage_option is not None):
            data_obj.to_file(path=os.path.join(conf.LOCAL_DATA_PATH, cache_name))
    else:
        ext_name = obj.split(".")[-1]
        if ext_name == "csv":
            data_obj = pd.read_csv(obj, storage_options=storage_option)
            if need_cache & (storage_option is not None):
                data_obj.to_csv(
                    path_or_buf=os.path.join(conf.LOCAL_DATA_PATH, cache_name)
                )
        elif ext_name == "txt":
            data_obj = pd.read_fwf(obj, storage_options=storage_option)
            if need_cache & (storage_option is not None):
                data_obj.to_csv(os.path.join(conf.LOCAL_DATA_PATH, cache_name))
        elif (ext_name in ["nc", "nc4", "h5", "hdf5"]) or ("nc4" in obj):
            if need_refer:
                data_obj = gen_refer_and_read_zarr(obj, storage_option=storage_option)
            else:
                """
                nc_source = intk.datatypes.HDF5(obj, storage_options=storage_option)
                nc_src_reader = intk.readers.DaskHDF(nc_source).to_reader()
                data_obj: xr.Dataset = nc_src_reader.read()
                """
                if storage_option is None:
                    if (ext_name == "nc") or (ext_name == "nc4") or ("nc4" in obj):
                        data_obj = xr.open_dataset(obj, chunks="auto")
                    elif ext_name in ["hdf5", "h5"]:
                        data_obj = xr.open_dataset(
                            obj, engine="h5netcdf", chunks="auto", phony_dims="access"
                        )
                elif (ext_name == "nc") or (ext_name == "nc4") or ("nc4" in obj):
                    data_obj = xr.open_dataset(conf.FS.open(obj), chunks="auto")
                elif ext_name in ["hdf5", "h5"]:
                    data_obj = xr.open_dataset(
                        conf.FS.open(obj),
                        engine="h5netcdf",
                        chunks="auto",
                        phony_dims="access",
                    )
            if (need_cache is True) & (storage_option is not None):
                data_obj.to_netcdf(path=os.path.join(conf.LOCAL_DATA_PATH, cache_name))
        elif ext_name == "json":
            data_obj = pd.read_json(obj, storage_options=storage_option)
            if (need_cache is True) & (storage_option is not None):
                data_obj.to_json(
                    path_or_buf=os.path.join(conf.LOCAL_DATA_PATH, cache_name)
                )
        elif ext_name == "zip":
            # Now zipfile is used to read shapefile
            # Can't read shapefile directly, see this: https://github.com/geopandas/geopandas/issues/3129
            # pip install pyogrio
            if storage_option is not None:
                data_obj = gpd.read_file(conf.FS.open(obj), engine="pyogrio")
            else:
                data_obj = gpd.read_file(obj)
            if (need_cache is True) & (storage_option is not None):
                data_obj.to_file(path=os.path.join(conf.LOCAL_DATA_PATH, cache_name))
        elif "grb2" in obj:
            if storage_option is not None:
                obj = f"simplecache::{obj}"
                grib_ds = xr.open_dataset(
                    fsspec.open_local(
                        obj,
                        s3=conf.MINIO_PARAM,
                        filecache={"cache_storage": "/tmp/files"},
                    ),
                    engine="cfgrib",
                )
            else:
                grib_ds = xr.open_dataset(obj)
            if (need_cache is True) & (storage_option is not None):
                grib_ds.to_netcdf(path=os.path.join(conf.LOCAL_DATA_PATH, cache_name))
        elif ext_name == "zarr":
            if storage_option is not None:
                zarr_mapper = conf.FS.get_mapper(obj)
                # KVStore is introduced in zarr specification V3
                # https://zarr-specs.readthedocs.io/en/latest/v3/stores.html
                zarr_store = zarr.storage.KVStore(zarr_mapper)
                data_obj = xr.open_zarr(zarr_store)
            else:
                data_obj = xr.open_zarr(obj)
        else:
            logging.error(f"Unsupported file type: {ext_name}")
    return data_obj


def gen_refer_and_read_zarr(obj_path, storage_option=None):
    # https://github.com/fsspec/kerchunk/discussions/431
    obj_json_path = f"s3://references/{str(obj_path)}.json"
    if not conf.FS.exists(obj_json_path):
        obj_path = "s3://" + obj_path
        with conf.FS.open(obj_path, "rb") as fpj:
            nc_chunks = SingleHdf5ToZarr(fpj, obj_path)
            with conf.FS.open(obj_json_path, "wb") as fp:
                fp.write(ujson.dumps(nc_chunks.translate()).encode())
    data_type_obj = intk.datatypes.HDF5(obj_json_path, storage_options=storage_option)
    data_reader_obj = intk.readers.XArrayDatasetReader(data_type_obj).to_reader()
    return data_reader_obj.read()



"""
该模块用于从minio中读取gpm数据，主要方法包括：

- `open_dataset` - 普通读取数据方法
- `from_shp` - 通过已有矢量范围读取数据方法
- `from_aoi` - 通过已有GeoDataFrame范围读取数据方法
"""

import numpy as np
import xarray as xr
import calendar
import dask
import json

from ..configs.config import FS, MINIO_PARAM

from ..configs.config import RO
from ..utils.utils import regen_box

bucket_name = MINIO_PARAM["bucket_name"]


# 后期从minio读取
start = np.datetime64("2016-01-01T00:00:00.000000000")
end = np.datetime64("2023-08-17T23:30:00.000000000")
change = np.datetime64("2023-07-01T23:30:00.000000000")

with FS.open(f"{bucket_name}/geodata/gpm/gpm.json") as f:
    cont = json.load(f)
end = np.datetime64(cont["end"])

box = (73.05, 3.05, 135.95, 53.95)

variables = [
    "HQobservationTime",
    "HQprecipSource",
    "HQprecipitation",
    "IRkalmanFilterWeight",
    "IRprecipitation",
    "precipitationCal",
    "precipitationUncal",
    "probabilityLiquidPrecipitation",
    "randomError",
]


dask.config.set({"array.slicing.split_large_chunks": False})


def get_dataset_year(start_time, end_time, bbox, time_chunks):
    year = str(start_time)[:4]

    chunks = {"time": time_chunks}
    ds = xr.open_dataset(
        "reference://",
        engine="zarr",
        chunks=chunks,
        backend_kwargs={
            "consolidated": False,
            "storage_options": {
                "fo": FS.open(
                    f"s3://{bucket_name}/geodata/gpm/{year}/gpm{year}_inc.json"
                ),
                "remote_protocol": "s3",
                "remote_options": RO,
            },
        },
    )

    ds = ds["precipitationCal"]
    # ds.to_dataframe().filter(['precipitationCal','precipitationUncal']).to_xarray()

    # ds = ds.rename({"longitude": "lon", "latitude": "lat"})
    ds = ds.transpose("time", "lon", "lat")

    if start_time < start:
        start_time = start

    if end_time > end:
        end_time = end

    times = slice(start_time, end_time)
    ds = ds.sel(time=times)

    bbox = regen_box(bbox, 0.1, 0.05)

    if bbox[0] < box[0]:
        left = box[0]
    else:
        left = bbox[0]

    if bbox[1] < box[1]:
        bottom = box[1]
    else:
        bottom = bbox[1]

    if bbox[2] > box[2]:
        right = box[2]
    else:
        right = bbox[2]

    if bbox[3] > box[3]:
        top = box[3]
    else:
        top = bbox[3]

    longitudes = slice(left - 0.00001, right + 0.00001)
    latitudes = slice(bottom - 0.00001, top + 0.00001)

    ds = ds.sortby("lat", ascending=True)
    ds = ds.sel(lon=longitudes, lat=latitudes)

    return ds


def get_dataset_month(start_time, end_time, bbox, time_chunks):
    year = str(start_time)[:4]
    month = str(start_time)[5:7].zfill(2)

    chunks = {"time": time_chunks}
    ds = xr.open_dataset(
        "reference://",
        engine="zarr",
        chunks=chunks,
        backend_kwargs={
            "consolidated": False,
            "storage_options": {
                "fo": FS.open(
                    f"s3://{bucket_name}/geodata/gpm/{year}/{month}/gpm{year}{month}_inc.json"
                ),
                "remote_protocol": "s3",
                "remote_options": RO,
            },
        },
    )

    ds = cf2datetime(ds)
    ds = ds["precipitationCal"]
    # ds.to_dataframe().filter(['precipitationCal','precipitationUncal']).to_xarray()

    # ds = ds.rename({"longitude": "lon", "latitude": "lat"})
    ds = ds.transpose("time", "lon", "lat")

    if start_time < start:
        start_time = start

    if end_time > end:
        end_time = end

    times = slice(start_time, end_time)
    ds = ds.sel(time=times)

    bbox = regen_box(bbox, 0.1, 0.05)

    if bbox[0] < box[0]:
        left = box[0]
    else:
        left = bbox[0]

    if bbox[1] < box[1]:
        bottom = box[1]
    else:
        bottom = bbox[1]

    if bbox[2] > box[2]:
        right = box[2]
    else:
        right = bbox[2]

    if bbox[3] > box[3]:
        top = box[3]
    else:
        top = bbox[3]

    longitudes = slice(left - 0.00001, right + 0.00001)
    latitudes = slice(bottom - 0.00001, top + 0.00001)

    ds = ds.sortby("lat", ascending=True)
    ds = ds.sel(lon=longitudes, lat=latitudes)

    return ds


def get_dataset_day(start_time, end_time, bbox, time_chunks):
    year = str(start_time)[:4]
    month = str(start_time)[5:7].zfill(2)
    day = str(end)[8:10].zfill(2)

    chunks = {"time": time_chunks}
    ds = xr.open_dataset(
        "reference://",
        engine="zarr",
        chunks=chunks,
        backend_kwargs={
            "consolidated": False,
            "storage_options": {
                "fo": FS.open(
                    f"s3://{bucket_name}/geodata/gpm/{year}/{month}/gpm{year}{month}_{day}.json"
                ),
                "remote_protocol": "s3",
                "remote_options": RO,
            },
        },
    )

    ds = cf2datetime(ds)
    ds = ds["precipitationCal"]
    # ds.to_dataframe().filter(['precipitationCal','precipitationUncal']).to_xarray()

    # ds = ds.rename({"longitude": "lon", "latitude": "lat"})
    ds = ds.transpose("time", "lon", "lat")

    if start_time < start:
        start_time = start

    if end_time > end:
        end_time = end

    times = slice(start_time, end_time)
    ds = ds.sel(time=times)

    bbox = regen_box(bbox, 0.1, 0.05)

    if bbox[0] < box[0]:
        left = box[0]
    else:
        left = bbox[0]

    if bbox[1] < box[1]:
        bottom = box[1]
    else:
        bottom = bbox[1]

    if bbox[2] > box[2]:
        right = box[2]
    else:
        right = bbox[2]

    if bbox[3] > box[3]:
        top = box[3]
    else:
        top = bbox[3]

    longitudes = slice(left - 0.00001, right + 0.00001)
    latitudes = slice(bottom - 0.00001, top + 0.00001)

    ds = ds.sortby("lat", ascending=True)
    ds = ds.sel(lon=longitudes, lat=latitudes)

    return ds


def cf2datetime(ds):
    ds = ds.copy()
    time_tmp1 = ds.indexes["time"]
    attrs = ds.coords["time"].attrs
    time_tmp2 = []
    for i in range(time_tmp1.shape[0]):
        tmp = time_tmp1[i]
        a = str(tmp.year).zfill(4)
        b = str(tmp.month).zfill(2)
        c = str(tmp.day).zfill(2)
        d = str(tmp.hour).zfill(2)
        e = str(tmp.minute).zfill(2)
        f = str(tmp.second).zfill(2)
        time_tmp2.append(
            np.datetime64("{}-{}-{} {}:{}:{}.00000000".format(a, b, c, d, e, f))
        )
    ds = ds.assign_coords(time=time_tmp2)
    ds.coords["time"].attrs = attrs

    return ds


def open_dataset(
    start_time=np.datetime64("2023-01-01T00:00:00.000000000"),
    end_time=np.datetime64("2023-01-02T00:00:00.000000000"),
    bbox=box,
    time_chunks=48,
):
    """
    从minio服务器读取gpm数据

    Args:
        start_time (datetime64): 开始时间
        end_time (datetime64): 结束时间
        bbox (list|tuple): 四至范围
        time_chunks (int): 分块数量

    Returns:
        dataset (Dataset): 读取结果
    """

    if end_time <= start_time:
        raise Exception("结束时间不能早于开始时间")

    if end_time <= change:
        # 早于20230701

        year_start = int(str(start_time)[:4])
        year_end = int(str(end_time)[:4])

        if year_start == year_end:
            ds = get_dataset_year(
                start_time=start_time,
                end_time=end_time,
                bbox=bbox,
                time_chunks=time_chunks,
            )
            return ds

        elif year_start < year_end:
            dss = []
            years = range(year_start, year_end + 1)
            for year in years:
                if year == year_start:
                    dss.append(
                        get_dataset_year(
                            start_time=start_time,
                            end_time=np.datetime64(f"{year}-12-31T23:30:00.000000000"),
                            bbox=bbox,
                            time_chunks=time_chunks,
                        )
                    )

                elif year == year_end:
                    dss.append(
                        get_dataset_year(
                            start_time=np.datetime64(
                                f"{year}-01-01T00:00:00.000000000"
                            ),
                            end_time=end_time,
                            bbox=bbox,
                            time_chunks=time_chunks,
                        )
                    )

                else:
                    dss.append(
                        get_dataset_year(
                            start_time=np.datetime64(
                                f"{year}-01-01T00:00:00.000000000"
                            ),
                            end_time=np.datetime64(f"{year}-12-31T23:30:00.000000000"),
                            bbox=bbox,
                            time_chunks=time_chunks,
                        )
                    )
            return xr.merge(dss)

    elif change <= start_time:
        # 晚于20230701

        year_start = int(str(start_time)[:4])
        year_end = int(str(end_time)[:4])
        month_start = int(str(start_time)[5:7])
        month_end = int(str(end_time)[5:7])
        end_month = int(str(end)[5:7])

        if year_start == year_end:
            if month_end < end_month:
                if month_start == month_end:
                    ds = get_dataset_month(
                        start_time=start_time,
                        end_time=end_time,
                        bbox=bbox,
                        time_chunks=time_chunks,
                    )
                    return ds

                else:
                    dss = []
                    for m in range(month_start, month_end + 1):
                        if m == month_start:
                            d = calendar.monthrange(year_start, m)[1]
                            dss.append(
                                get_dataset_month(
                                    start_time=start_time,
                                    end_time=np.datetime64(
                                        f"{year_start}-{str(m).zfill(2)}-{str(d).zfill(2)}T23:30:00.000000000"
                                    ),
                                    bbox=bbox,
                                    time_chunks=time_chunks,
                                )
                            )
                        elif m == month_end:
                            dss.append(
                                get_dataset_month(
                                    start_time=np.datetime64(
                                        f"{year_start}-{str(m).zfill(2)}-01T00:00:00.000000000"
                                    ),
                                    end_time=end_time,
                                    bbox=bbox,
                                    time_chunks=time_chunks,
                                )
                            )
                        else:
                            d = calendar.monthrange(year_start, m)[1]
                            dss.append(
                                get_dataset_month(
                                    start_time=np.datetime64(
                                        f"{year_start}-{str(m).zfill(2)}-01T00:00:00.000000000"
                                    ),
                                    end_time=np.datetime64(
                                        f"{year_start}-{str(m).zfill(2)}-{str(d).zfill(2)}T23:30:00.000000000"
                                    ),
                                    bbox=bbox,
                                    time_chunks=time_chunks,
                                )
                            )

                    return xr.merge(dss)

            else:
                if month_start == month_end:
                    ds = get_dataset_day(
                        start_time=start_time,
                        end_time=end_time,
                        bbox=bbox,
                        time_chunks=time_chunks,
                    )
                    return ds

                else:
                    dss = []
                    for m in range(month_start, month_end + 1):
                        if m == month_start:
                            d = calendar.monthrange(year_start, m)[1]
                            dss.append(
                                get_dataset_month(
                                    start_time=start_time,
                                    end_time=np.datetime64(
                                        f"{year_start}-{str(m).zfill(2)}-{str(d).zfill(2)}T23:30:00.000000000"
                                    ),
                                    bbox=bbox,
                                    time_chunks=time_chunks,
                                )
                            )
                        elif m == month_end:
                            dss.append(
                                get_dataset_day(
                                    start_time=np.datetime64(
                                        f"{year_start}-{str(m).zfill(2)}-01T00:00:00.000000000"
                                    ),
                                    end_time=end_time,
                                    bbox=bbox,
                                    time_chunks=time_chunks,
                                )
                            )
                        else:
                            d = calendar.monthrange(year_start, m)[1]
                            dss.append(
                                get_dataset_month(
                                    start_time=np.datetime64(
                                        f"{year_start}-{str(m).zfill(2)}-01T00:00:00.000000000"
                                    ),
                                    end_time=np.datetime64(
                                        f"{year_start}-{str(m).zfill(2)}-{str(d).zfill(2)}T23:30:00.000000000"
                                    ),
                                    bbox=bbox,
                                    time_chunks=time_chunks,
                                )
                            )

                    return xr.merge(dss)

        else:
            dss = []

            for y in range(year_start, year_end):
                if y == year_start:
                    for m in range(month_start, 13):
                        if m == month_start:
                            d = calendar.monthrange(y, m)[1]
                            dss.append(
                                get_dataset_month(
                                    start_time=start_time,
                                    end_time=np.datetime64(
                                        f"{str(y).zfill(4)}-{str(m).zfill(2)}-{str(d).zfill(2)}T23:30:00.000000000"
                                    ),
                                    bbox=bbox,
                                    time_chunks=time_chunks,
                                )
                            )
                        else:
                            d = calendar.monthrange(y, m)[1]
                            dss.append(
                                get_dataset_month(
                                    start_time=np.datetime64(
                                        f"{str(y).zfill(4)}-{str(m).zfill(2)}-01T00:00:00.000000000"
                                    ),
                                    end_time=np.datetime64(
                                        f"{str(y).zfill(4)}-{str(m).zfill(2)}-{str(d).zfill(2)}T23:30:00.000000000"
                                    ),
                                    bbox=bbox,
                                    time_chunks=time_chunks,
                                )
                            )

                elif y == year_end:
                    for m in range(1, month_end + 1):
                        if m == end_month:
                            dss.append(
                                get_dataset_day(
                                    start_time=np.datetime64(
                                        f"{str(y).zfill(4)}-{str(m).zfill(2)}-01T00:00:00.000000000"
                                    ),
                                    end_time=end_time,
                                    bbox=bbox,
                                    time_chunks=time_chunks,
                                )
                            )
                        else:
                            dss.append(
                                get_dataset_month(
                                    start_time=np.datetime64(
                                        f"{str(y).zfill(4)}-{str(m).zfill(2)}-01T00:00:00.000000000"
                                    ),
                                    end_time=np.datetime64(
                                        f"{str(y).zfill(4)}-{str(m).zfill(2)}-{str(d).zfill(2)}T23:30:00.000000000"
                                    ),
                                    bbox=bbox,
                                    time_chunks=time_chunks,
                                )
                            )

                else:
                    for m in range(1, 13):
                        d = calendar.monthrange(y, m)[1]
                        dss.append(
                            get_dataset_month(
                                start_time=np.datetime64(
                                    f"{str(y).zfill(4)}-{str(m).zfill(2)}-01T00:00:00.000000000"
                                ),
                                end_time=np.datetime64(
                                    f"{str(y).zfill(4)}-{str(m).zfill(2)}-{str(d).zfill(2)}T23:30:00.000000000"
                                ),
                                bbox=bbox,
                                time_chunks=time_chunks,
                            )
                        )

            return xr.merge(dss)

    else:
        # 包含20230701

        year_start = int(str(start_time)[:4])
        year_end = int(str(end_time)[:4])
        month_start = int(str(start_time)[5:7])
        month_end = int(str(end_time)[5:7])
        end_month = int(str(end)[5:7])

        dss = []

        for y in range(year_start, 2024):
            if y == year_start and y == 2023:
                dss.append(
                    get_dataset_year(
                        start_time=start_time,
                        end_time=np.datetime64("2023-07-01T23:30:00.000000000"),
                        bbox=bbox,
                        time_chunks=time_chunks,
                    )
                )
            elif y == year_start and y < 2023:
                dss.append(
                    get_dataset_year(
                        start_time=start_time,
                        end_time=np.datetime64("2023-12-31T23:30:00.000000000"),
                        bbox=bbox,
                        time_chunks=time_chunks,
                    )
                )
            elif y == 2023:
                dss.append(
                    get_dataset_year(
                        start_time=np.datetime64("2023-01-01T00:00:00.000000000"),
                        end_time=np.datetime64("2023-07-01T23:30:00.000000000"),
                        bbox=bbox,
                        time_chunks=time_chunks,
                    )
                )
            else:
                dss.append(
                    get_dataset_year(
                        start_time=np.datetime64("2023-01-01T00:00:00.000000000"),
                        end_time=np.datetime64("2023-12-31T23:30:00.000000000"),
                        bbox=bbox,
                        time_chunks=time_chunks,
                    )
                )

        for y in range(2023, year_end + 1):
            if y == year_end and y == 2023:
                for m in range(7, month_end + 1):
                    if m == month_end and m == 7:
                        dss.append(
                            get_dataset_month(
                                start_time=np.datetime64(
                                    f"{str(y).zfill(4)}-{str(m).zfill(2)}-02T00:00:00.000000000"
                                ),
                                end_time=end_time,
                                bbox=bbox,
                                time_chunks=time_chunks,
                            )
                        )
                    elif m == month_end and m == end_month:
                        dss.append(
                            get_dataset_day(
                                start_time=np.datetime64(
                                    f"{str(y).zfill(4)}-{str(m).zfill(2)}-01T00:00:00.000000000"
                                ),
                                end_time=end_time,
                                bbox=bbox,
                                time_chunks=time_chunks,
                            )
                        )
                    elif m == month_end and m > 7:
                        dss.append(
                            get_dataset_month(
                                start_time=np.datetime64(
                                    f"{str(y).zfill(4)}-{str(m).zfill(2)}-01T00:00:00.000000000"
                                ),
                                end_time=end_time,
                                bbox=bbox,
                                time_chunks=time_chunks,
                            )
                        )
                    elif m == 7:
                        dss.append(
                            get_dataset_month(
                                start_time=np.datetime64(
                                    "2023-07-02T00:00:00.000000000"
                                ),
                                end_time=np.datetime64("2023-07-31T23:30:00.000000000"),
                                bbox=bbox,
                                time_chunks=time_chunks,
                            )
                        )
                    else:
                        d = calendar.monthrange(y, m)[1]
                        dss.append(
                            get_dataset_year(
                                start_time=np.datetime64(
                                    f"{str(y).zfill(4)}-{str(m).zfill(2)}-01T00:00:00.000000000"
                                ),
                                end_time=np.datetime64(
                                    f"{str(y).zfill(4)}-{str(m).zfill(2)}-{str(d).zfill(2)}T23:30:00.000000000"
                                ),
                                bbox=bbox,
                                time_chunks=time_chunks,
                            )
                        )
            elif y == year_end and y > 2023:
                for m in range(1, month_end + 1):
                    if m == month_end and m == end_month:
                        dss.append(
                            get_dataset_day(
                                start_time=np.datetime64(
                                    f"{str(y).zfill(4)}-{str(m).zfill(2)}-01T00:00:00.000000000"
                                ),
                                end_time=end_time,
                                bbox=bbox,
                                time_chunks=time_chunks,
                            )
                        )
                    elif m == month_end:
                        dss.append(
                            get_dataset_month(
                                start_time=np.datetime64(
                                    f"{str(y).zfill(4)}-{str(m).zfill(2)}-01T00:00:00.000000000"
                                ),
                                end_time=end_time,
                                bbox=bbox,
                                time_chunks=time_chunks,
                            )
                        )
                    else:
                        d = calendar.monthrange(y, m)[1]
                        dss.append(
                            get_dataset_year(
                                start_time=np.datetime64(
                                    f"{str(y).zfill(4)}-{str(m).zfill(2)}-01T00:00:00.000000000"
                                ),
                                end_time=np.datetime64(
                                    f"{str(y).zfill(4)}-{str(m).zfill(2)}-{str(d).zfill(2)}T23:30:00.000000000"
                                ),
                                bbox=bbox,
                                time_chunks=time_chunks,
                            )
                        )
            elif y == 2023:
                for m in range(7, 13):
                    if m == 7:
                        dss.append(
                            get_dataset_year(
                                start_time=np.datetime64(
                                    "2023-07-02T00:00:00.000000000"
                                ),
                                end_time=np.datetime64("2023-07-31T23:30:00.000000000"),
                                bbox=bbox,
                                time_chunks=time_chunks,
                            )
                        )
                    else:
                        d = calendar.monthrange(y, m)[1]
                        dss.append(
                            get_dataset_year(
                                start_time=np.datetime64(
                                    f"{str(y).zfill(4)}-{str(m).zfill(2)}-01T00:00:00.000000000"
                                ),
                                end_time=np.datetime64(
                                    f"{str(y).zfill(4)}-{str(m).zfill(2)}-{str(d).zfill(2)}T23:30:00.000000000"
                                ),
                                bbox=bbox,
                                time_chunks=time_chunks,
                            )
                        )
            else:
                for m in range(1, 13):
                    d = calendar.monthrange(y, m)[1]
                    dss.append(
                        get_dataset_year(
                            start_time=np.datetime64(
                                f"{str(y).zfill(4)}-{str(m).zfill(2)}-01T00:00:00.000000000"
                            ),
                            end_time=np.datetime64(
                                f"{str(y).zfill(4)}-{str(m).zfill(2)}-{str(d).zfill(2)}T23:30:00.000000000"
                            ),
                            bbox=bbox,
                            time_chunks=time_chunks,
                        )
                    )

        return xr.merge(dss)


import geopandas as gpd


def from_shp(
    start_time=np.datetime64("2023-01-01T00:00:00.000000000"),
    end_time=np.datetime64("2023-01-02T00:00:00.000000000"),
    shp=None,
    time_chunks=48,
):
    """
    通过已有的矢量数据范围从minio服务器读取gpm数据

    Args:
        start_time (datetime64): 开始时间
        end_time (datetime64): 结束时间
        shp (str): 矢量数据路径
        time_chunks (int): 分块数量

    Returns:
        dataset (Dataset): 读取结果
    """

    gdf = gpd.GeoDataFrame.from_file(shp)
    b = gdf.bounds
    bbox = regen_box(
        (b.loc[0]["minx"], b.loc[0]["miny"], b.loc[0]["maxx"], b.loc[0]["maxy"]),
        0.1,
        0.05,
    )

    ds = open_dataset(start_time, end_time, bbox, time_chunks)

    return ds


def from_aoi(
    start_time=np.datetime64("2023-01-01T00:00:00.000000000"),
    end_time=np.datetime64("2023-01-02T00:00:00.000000000"),
    aoi: gpd.GeoDataFrame = None,
    time_chunks=48,
):
    """
    用过已有的GeoPandas.GeoDataFrame对象从minio服务器读取gpm数据

    Args:
        start_time (datetime64): 开始时间
        end_time (datetime64): 结束时间
        aoi (GeoDataFrame): 已有的GeoPandas.GeoDataFrame对象
        time_chunks (int): 分块数量

    Returns:
        dataset (Dataset): 读取结果
    """

    b = aoi.bounds
    bbox = regen_box(
        (b.loc[0]["minx"], b.loc[0]["miny"], b.loc[0]["maxx"], b.loc[0]["maxy"]),
        0.1,
        0.05,
    )

    ds = open_dataset(start_time, end_time, bbox, time_chunks)

    return ds


if __name__ == "__main__":
    pass



"""
Author: Wenyu Ouyang
Date: 2023-10-25 17:12:30
LastEditTime: 2024-03-28 08:33:13
LastEditors: Wenyu Ouyang
Description: Preprocess data source -- transform its format to tidy format
FilePath: \hydrodatasource\hydrodatasource\reader\stations.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

import logging
import os
import pandas as pd

# from hydroutils import hydro_logger
from hydrodatasource.configs.config import (
    LOCAL_DATA_PATH,
    MC,
    STATION_BUCKET,
    STATION_OBJECT,
)
from hydrodatasource.reader.minio_api import minio_upload_csv


def huanren_preprocess():
    """
    Process data from huanren
    """
    # read local data
    try:
        huanren_df = pd.read_excel(
            os.path.join(LOCAL_DATA_PATH, "huanren", "桓仁多年逐日入库流量.xls"),
            sheet_name=1,
        )
        tidy_df2 = convert_to_tidy(huanren_df, "format2")
        local_save_path = os.path.join(LOCAL_DATA_PATH, "huanren", "tidy.csv")
        tidy_df2.to_csv(local_save_path, index=False)
        # boto3_upload_csv(s3, site_bucket, site_object, local_save_path)
        minio_upload_csv(MC, STATION_BUCKET, STATION_OBJECT, local_save_path)
    except Exception as e:
        preview = str(e)
        # hydro_logger.error(preview)
        logging.error(preview)


def biliu_stbprp_decode():
    biliu_stbprp_path = os.path.join(LOCAL_DATA_PATH, "biliu", "st_stbprp_b.xls")
    biliu_df = pd.read_excel(biliu_stbprp_path, sheet_name="st_stbprp_b")
    new_stids = []
    for stid in biliu_df["stid"]:
        new_stid = "2145" + str(stid).rjust(3, "0") + "X"
        new_stids.append(new_stid)
    biliu_df["stid"] = new_stids
    minio_upload_csv(
        MC, STATION_BUCKET, "test_stations/st_stbprp_b.xls", file_path=biliu_stbprp_path
    )



import geopandas as gpd
import numpy as np
import pandas as pd
from geopandas import GeoDataFrame
from scipy.spatial import Voronoi
from shapely.geometry import Polygon
import hydrodatasource.configs.config as hdscc


def read_data(rainfall_data_paths: list, head='local', check_time=None):
    # Read rainfall CSV files
    rainfall_dfs = []
    check_time = pd.to_datetime(check_time, format='%Y-%m-%d %H:%M:%S')
    latest_date = pd.Timestamp.min  # initialize latest date as minimum Timestamp
    # Find latest date in CSV files
    for file in rainfall_data_paths:
        if head == 'local':
            df = pd.read_csv(file)
        elif head == 'minio':
            df = pd.read_csv(file, storage_options=hdscc.MINIO_PARAM)
        else:
            df = pd.DataFrame()
        first_row_date = pd.to_datetime(df.iloc[0]['TM'])
        if (first_row_date > latest_date) & (first_row_date <= check_time):
            latest_date = first_row_date
            rainfall_dfs.append(df)
    # Convert rainfall data and filter by latest date
    if len(rainfall_dfs) > 0:
        rainfall_df = pd.concat(rainfall_dfs).drop_duplicates().reset_index(drop=True)
        rainfall_df['TM'] = pd.to_datetime(rainfall_df['TM'])
        rainfall_df = rainfall_df[rainfall_df['TM'] >= latest_date]
    else:
        temp_range = pd.date_range('1990-01-01', '2038-12-31', freq='h')
        rainfall_df = pd.DataFrame({'TM': temp_range, 'DRP': np.repeat(0, len(temp_range.to_list()))})
    return rainfall_df


def calculate_voronoi_polygons(stations, basin_geom):
    """Calculate Voronoi polygons for each station."""
    bounding_box = basin_geom.envelope.exterior.coords
    points = np.array([point.coords[0] for point in stations.geometry])
    points_extended = np.concatenate((points, bounding_box))
    vor = Voronoi(points_extended)
    regions = [vor.regions[vor.point_region[i]] for i in range(len(points))]
    polygons = [Polygon(vor.vertices[region]).buffer(0) for region in regions if -1 not in region]
    polygons_gdf = gpd.GeoDataFrame(geometry=polygons, crs=stations.crs)
    polygons_gdf['station_id'] = stations['STCD'].astype(str).values
    polygons_gdf['original_area'] = polygons_gdf.geometry.area
    clipped_polygons_gdf = gpd.clip(polygons_gdf, basin_geom)
    clipped_polygons_gdf['clipped_area'] = clipped_polygons_gdf.geometry.area
    total_basin_area = basin_geom.area
    clipped_polygons_gdf['area_ratio'] = clipped_polygons_gdf['clipped_area'] / total_basin_area
    return clipped_polygons_gdf


def calculate_weighted_rainfall(voronoi_polygons, rainfall_data):
    voronoi_polygons['station_id'] = voronoi_polygons['station_id'].astype(str)
    if 'STCD' in rainfall_data.columns:
        rainfall_data['station_id'] = rainfall_data['STCD'].astype(str)
    else:
        rainfall_data['station_id'] = voronoi_polygons['station_id'].astype(str)
    merged_data = pd.merge(voronoi_polygons, rainfall_data, on='station_id')
    merged_data['weighted_rainfall'] = merged_data['DRP'] * merged_data['area_ratio']
    weighted_average_rainfall = merged_data.groupby('TM')['weighted_rainfall'].sum()
    return weighted_average_rainfall.reset_index()


def rainfall_average(basin: GeoDataFrame, stations_gdf: GeoDataFrame, pp_ids: list, check_time):
    # 如果新站对不上老时间，就会把老数据也砍掉
    check_time_path = f"{basin['BASIN_ID'][0]}_{check_time}_rainfall_mean.csv"
    s3_check_time_path = 's3://basins-origin/hour_data/1h/mean_data/' + check_time_path
    if hdscc.FS.exists(s3_check_time_path):
        return pd.read_csv(s3_check_time_path, storage_options=hdscc.MINIO_PARAM)
    else:
        rainfall_data_paths = [f's3://stations-origin/pp_stations/hour_data/1h/pp_{ppid}.csv' for ppid in pp_ids]
        rainfall_df = read_data(rainfall_data_paths, head='minio', check_time=check_time)
        stations_within_basin = gpd.sjoin(stations_gdf, basin)
        voronoi_polygons = calculate_voronoi_polygons(stations_within_basin, basin.geometry[0])
        average_rainfall = calculate_weighted_rainfall(voronoi_polygons, rainfall_df)
        # basin_name = os.path.splitext(os.path.basename(basin_shp_path))[0]
        # average_rainfall['basin_name'] = basin_name
        average_rainfall.to_csv(s3_check_time_path, index=False, storage_options=hdscc.MINIO_PARAM)
    return average_rainfall


'''
def plot_voronoi_polygons(original_polygons, clipped_polygons, basin):
    fig, (ax_original, ax_clipped) = plt.subplots(1, 2, figsize=(12, 6))
    original_polygons.plot(ax=ax_original, edgecolor='black')
    basin.boundary.plot(ax=ax_original, color='red')
    ax_original.set_title('Original Voronoi Polygons')
    clipped_polygons.plot(ax=ax_clipped, edgecolor='black')
    basin.boundary.plot(ax=ax_clipped, color='red')
    ax_clipped.set_title('Clipped Voronoi Polygons')
    plt.tight_layout()
    plt.show()
'''



"""
该模块用于从minio中读取gfs数据，主要方法包括：

- `open_dataset` - 普通读取数据方法
- `from_shp` - 通过已有矢量范围读取数据方法
- `from_aoi` - 通过已有GeoDataFrame范围读取数据方法
"""

import os
import numpy as np
import s3fs
import xarray as xr
import calendar
from datetime import date
import json

from ..configs.config import FS, MINIO_PARAM

from ..configs.config import RO
from ..utils.utils import regen_box

bucket_name = MINIO_PARAM["bucket_name"]

# 后期从minio读取
start = np.datetime64("2016-07-10")
end = np.datetime64("2023-08-17")
change = np.datetime64("2022-09-01")
# change_date = date(2022,9,1)
box = (115, 38, 136, 54)


# 2t  ----- temperature_2m_above_ground
# 2sh ----- specific_humidity_2m_above_ground
# 2r  ----- relative_humidity_2m_above_ground
# 10u ----- u_component_of_wind_10m_above_ground
# 10v ----- v_component_of_wind_10m_above_ground
# tp  ----- total_precipitation_surface
# pwat----- precipitable_water_entire_atmosphere
# tcc ----- total_cloud_cover_entire_atmosphere
# dswrf ----- downward_shortwave_radiation_flux

variables = {
    "dswrf": "downward_shortwave_radiation_flux",
    "pwat": "precipitable_water_entire_atmosphere",
    "2r": "relative_humidity_2m_above_ground",
    "2sh": "specific_humidity_2m_above_ground",
    "2t": "temperature_2m_above_ground",
    "tcc": "total_cloud_cover_entire_atmosphere",
    "tp": "total_precipitation_surface",
    "10u": "u_component_of_wind_10m_above_ground",
    "10v": "v_component_of_wind_10m_above_ground",
}


def open_dataset(
    data_variable="tp",
    creation_date=np.datetime64("2022-09-01"),
    creation_time="00",
    bbox=box,
    time_chunks=24,
):
    """
    从minio服务器读取gfs数据

    Args:
        data_variables (str): 数据变量，目前只支持tp，即降雨
        creation_date (datetime64): 创建日期
        creation_time (datetime64): 创建时间，即00\06\12\18之一
        bbox (list|tuple): 四至范围
        time_chunks (int): 分块数量

    Returns:
        dataset (Dataset): 读取结果
    """

    if data_variable in variables.keys():
        short_name = data_variable
        full_name = variables[data_variable]

        with FS.open(f"{bucket_name}/geodata/gfs/gfs.json") as f:
            cont = json.load(f)
            start = np.datetime64(cont[short_name][0]["start"])
            end = np.datetime64(cont[short_name][-1]["end"])

        if creation_date < start or creation_date > end:
            print("超出时间范围！")
            return

        if creation_time not in ["00", "06", "12", "18"]:
            print("creation_time必须是00、06、12、18之一！")
            return

        year = str(creation_date.astype("object").year)
        month = str(creation_date.astype("object").month).zfill(2)
        day = str(creation_date.astype("object").day).zfill(2)

        if creation_date < change:
            json_url = f"s3://{bucket_name}/geodata/gfs/gfs_history/{year}/{month}/{day}/gfs{year}{month}{day}.t{creation_time}z.0p25.json"
        else:
            json_url = f"s3://{bucket_name}/geodata/gfs/{short_name}/{year}/{month}/{day}/gfs{year}{month}{day}.t{creation_time}z.0p25.json"

        chunks = {"valid_time": time_chunks}
        ds = xr.open_dataset(
            "reference://",
            engine="zarr",
            chunks=chunks,
            backend_kwargs={
                "consolidated": False,
                "storage_options": {
                    "fo": FS.open(json_url),
                    "remote_protocol": "s3",
                    "remote_options": RO,
                },
            },
        )

        if creation_date < change:
            ds = ds[full_name]

        # ds = ds.filter_by_attrs(long_name=lambda v: v in data_variables)
        ds = ds.rename({"longitude": "lon", "latitude": "lat"})
        # ds = ds.transpose('time','valid_time','lon','lat')

        bbox = regen_box(bbox, 0.25, 0)

        if bbox[0] < box[0]:
            left = box[0]
        else:
            left = bbox[0]

        if bbox[1] < box[1]:
            bottom = box[1]
        else:
            bottom = bbox[1]

        if bbox[2] > box[2]:
            right = box[2]
        else:
            right = bbox[2]

        if bbox[3] > box[3]:
            top = box[3]
        else:
            top = bbox[3]

        longitudes = slice(left - 0.00001, right + 0.00001)
        latitudes = slice(bottom - 0.00001, top + 0.00001)

        ds = ds.sortby("lat", ascending=True)
        ds = ds.sel(lon=longitudes, lat=latitudes)

        return ds

    else:
        print("变量名不存在！")


import geopandas as gpd


def from_shp(
    data_variable="tp",
    creation_date=np.datetime64("2022-09-01"),
    creation_time="00",
    shp=None,
    time_chunks=24,
):
    """
    通过已有的矢量数据范围从minio服务器读取gfs数据

    Args:
        data_variables (str): 数据变量，目前只支持tp，即降雨
        creation_date (datetime64): 创建日期
        creation_time (datetime64): 创建时间，即00\06\12\18之一
        shp (str): 矢量数据路径
        time_chunks (int): 分块数量

    Returns:
        dataset (Dataset): 读取结果
    """

    gdf = gpd.GeoDataFrame.from_file(shp)
    b = gdf.bounds
    bbox = regen_box(
        (b.loc[0]["minx"], b.loc[0]["miny"], b.loc[0]["maxx"], b.loc[0]["maxy"]), 0.1, 0
    )

    return open_dataset(data_variable, creation_date, creation_time, bbox, time_chunks)


def from_aoi(
    data_variable="tp",
    creation_date=np.datetime64("2022-09-01"),
    creation_time="00",
    aoi: gpd.GeoDataFrame = None,
    time_chunks=24,
):
    """
    通过已有的GeoPandas.GeoDataFrame对象从minio服务器读取gfs数据

    Args:
        data_variables (str): 数据变量，目前只支持tp，即降雨
        creation_date (datetime64): 创建日期
        creation_time (datetime64): 创建时间，即00\06\12\18之一
        aoi (GeoDataFrame): 已有的GeoPandas.GeoDataFrame对象
        time_chunks (int): 分块数量

    Returns:
        dataset (Dataset): 读取结果
    """
    b = aoi.bounds
    bbox = regen_box(
        (b.loc[0]["minx"], b.loc[0]["miny"], b.loc[0]["maxx"], b.loc[0]["maxy"]), 0.1, 0
    )

    return open_dataset(data_variable, creation_date, creation_time, bbox, time_chunks)



import os
import pandas as pd

from hydrodatasource.processor.data_checker import DataChecker


def test_check_folder_structure():
    """
    Test to check if the folder structure in the local data path matches the expected format.
    """
    # Initialize the DataChecker with the local data path from the settings
    data_checker = DataChecker()

    # Check the folder structure
    result = data_checker.check_folder_structure()

    # Assert that the result is correct
    assert result


def test_check_file_format(tmpdir):
    """
    Test to check if the file format is correct.
    """
    # Initialize the DataChecker with the local data path from the settings
    data_checker = DataChecker()

    # Define a sample file path and expected columns
    file_path = os.path.join(str(tmpdir), "file.csv")
    expected_columns = ["column1", "column2", "column3"]

    # Create a sample file with the expected columns
    sample_data = pd.DataFrame(columns=expected_columns)
    sample_data.to_csv(file_path, index=False)

    # Check the file format
    result = data_checker.check_file_format(file_path, expected_columns)

    # Assert that the result is True
    assert result

    # Clean up the sample file
    os.remove(file_path)


def test_check_files_in_folder(tmpdir):
    """
    Test to check the format of all CSV files in a folder.
    """
    # Initialize the DataChecker with the local data path from the settings
    data_checker = DataChecker()

    # Define a sample folder path and expected columns
    folder_path = str(tmpdir)
    expected_columns = ["column1", "column2", "column3"]

    # Create sample CSV files with the expected columns
    file1_path = os.path.join(folder_path, "file1.csv")
    file2_path = os.path.join(folder_path, "file2.csv")
    file3_path = os.path.join(folder_path, "file3.csv")

    sample_data = pd.DataFrame(columns=expected_columns)
    sample_data.to_csv(file1_path, index=False)
    sample_data.to_csv(file2_path, index=False)
    sample_data.to_csv(file3_path, index=False)

    # Check the file formats in the folder
    result = data_checker.check_files_in_folder(folder_path, expected_columns)

    # Assert that the result is True
    assert result

    # Clean up the sample files
    os.remove(file1_path)
    os.remove(file2_path)
    os.remove(file3_path)


def test_check_station_data_files(tmpdir):
    """
    Test to check the format of station data files.
    """
    # Initialize the DataChecker
    data_checker = DataChecker()

    # Define a sample folder path and expected columns
    folder_path = tmpdir.mkdir("stations-origin")
    expected_columns = {
        "pp_stations": ["column1", "column2", "column3"],
        "zz_stations": ["column4", "column5", "column6"],
        "zq_stations": ["column7", "column8", "column9"],
    }

    # Create directories and sample files for each station type
    for station_type in ["pp", "zz", "zq"]:
        # Create station directories
        station_dir = folder_path.mkdir(f"{station_type}_stations")
        basic_info_file = station_dir.join(f"{station_type}_stations.csv")
        time_series_file = station_dir.join("time_series.csv")

        # Create a sample file with the expected columns
        sample_data = pd.DataFrame(columns=expected_columns[f"{station_type}_stations"])
        sample_data.to_csv(str(basic_info_file), index=False)
        sample_data.to_csv(str(time_series_file), index=False)

    # Check the format of station data files
    result = data_checker.check_station_data_files()

    # Assert that the result is True
    assert result



import os

import geopandas as gpd

import hydrodatasource.configs.config as hdscc
from hydrodatasource.processor.basin_mean_rainfall import calculate_voronoi_polygons, calculate_weighted_rainfall, \
    read_data


def test_rainfall_average():
    basin_shp_path = 's3://basins-origin/basin_shapefiles/basin_CHN_songliao_21401550.zip'
    rainfall_stations = 's3://stations-origin/stations_list/pp_stations.zip'
    pp_ids = ['pp_CHN_biliuhezijian_4002.csv', 'pp_CHN_biliuhezijian_4003.csv', 'pp_CHN_biliuhezijian_4006.csv',
              'pp_CHN_biliuhezijian_4007.csv', 'pp_CHN_biliuhezijian_4008.csv']
    rainfall_data_paths = ['s3://stations-origin/pp_stations/hour_data/1h/' + ppid for ppid in pp_ids]
    rainfall_df = read_data(rainfall_data_paths, head='minio')
    basin = gpd.read_file(hdscc.FS.open(basin_shp_path))
    stations_gdf = gpd.read_file(hdscc.FS.open(rainfall_stations))
    stations_within_basin = gpd.sjoin(stations_gdf, basin)
    voronoi_polygons = calculate_voronoi_polygons(stations_within_basin, basin.geometry[0])
    average_rainfall = calculate_weighted_rainfall(voronoi_polygons, rainfall_df)
    basin_name = os.path.splitext(os.path.basename(basin_shp_path))[0]
    average_rainfall['basin_name'] = basin_name
    return average_rainfall



"""
Author: Jianfeng Zhu
Date: 2023-10-25 18:49:02
LastEditTime: 2024-08-15 10:02:10
LastEditors: Wenyu Ouyang
Description: Some configs for minio server
FilePath: \hydrodatasource\hydrodatasource\configs\config.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

import os
from pathlib import Path
import boto3
import s3fs
import yaml
from minio import Minio
import psycopg2


def read_setting(setting_path):
    if not os.path.exists(setting_path):
        raise FileNotFoundError(f"Configuration file not found: {setting_path}")

    with open(setting_path, "r", encoding="utf-8") as file:  # 指定编码为 UTF-8
        setting = yaml.safe_load(file)

    example_setting = (
        "minio:\n"
        "  server_url: 'http://minio.waterism.com:9090' # Update with your URL\n"
        "  client_endpoint: 'http://minio.waterism.com:9000' # Update with your URL\n"
        "  access_key: 'your minio access key'\n"
        "  secret: 'your minio secret'\n\n"
        "local_data_path:\n"
        "  root: 'D:\\data\\waterism' # Update with your root data directory\n"
        "  datasets-origin: 'D:\\data\\waterism\\datasets-origin'\n"
        "  datasets-interim: 'D:\\data\\waterism\\datasets-interim'\n"
        "postgres:\n"
        "  server_url: your_postgres_server_url\n"
        "  port: 5432\n"
        "  username: your_postgres_username\n"
        "  password: your_postgres_secret_code\n"
        "  database: your_postgres_database\n"
    )

    if setting is None:
        raise ValueError(
            f"Configuration file is empty or has invalid format.\n\nExample configuration:\n{example_setting}"
        )

    # Define the expected structure
    expected_structure = {
        "minio": ["server_url", "client_endpoint", "access_key", "secret"],
        "local_data_path": ["root", "datasets-origin", "datasets-interim"],
        "postgres": ["server_url", "port", "username", "password", "database"],
    }

    # Validate the structure
    try:
        for key, subkeys in expected_structure.items():
            if key not in setting:
                raise KeyError(f"Missing required key in config: {key}")

            if isinstance(subkeys, list):
                for subkey in subkeys:
                    if subkey not in setting[key]:
                        raise KeyError(f"Missing required subkey '{subkey}' in '{key}'")
    except KeyError as e:
        raise ValueError(
            f"Incorrect configuration format: {e}\n\nExample configuration:\n{example_setting}"
        ) from e

    return setting


SETTING_FILE = os.path.join(Path.home(), "hydro_setting.yml")
try:
    SETTING = read_setting(SETTING_FILE)
except ValueError as e:
    print(e)
except Exception as e:
    print(f"Unexpected error: {e}")

# Initialize local data path
LOCAL_DATA_PATH = SETTING["local_data_path"]["root"]


# Initialize remote service settings
MINIO_PARAM = {}
RO = {}
S3 = None
MC = None
PS = None
FS = None

STATION_BUCKET = "stations"
STATION_OBJECT = "sites.csv"
GRID_INTERIM_BUCKET = "grids-interim"

# Handle MinIO connection
try:
    # MinIO parameters
    MINIO_PARAM = {
        "endpoint_url": SETTING["minio"]["client_endpoint"],
        "key": SETTING["minio"]["access_key"],
        "secret": SETTING["minio"]["secret"],
    }

    # Initialize S3 FileSystem
    FS = s3fs.S3FileSystem(
        client_kwargs={"endpoint_url": MINIO_PARAM["endpoint_url"]},
        key=MINIO_PARAM["key"],
        secret=MINIO_PARAM["secret"],
        use_ssl=False,
    )

    # remote_options parameters for xr open_dataset from minio
    RO = {
        "client_kwargs": {"endpoint_url": MINIO_PARAM["endpoint_url"]},
        "key": MINIO_PARAM["key"],
        "secret": MINIO_PARAM["secret"],
        "use_ssl": False,
    }

    # Set up MinIO client
    S3 = boto3.client(
        "s3",
        endpoint_url=SETTING["minio"]["client_endpoint"],
        aws_access_key_id=MINIO_PARAM["key"],
        aws_secret_access_key=MINIO_PARAM["secret"],
    )
    MC = Minio(
        SETTING["minio"]["client_endpoint"].replace("http://", ""),
        access_key=MINIO_PARAM["key"],
        secret_key=MINIO_PARAM["secret"],
        secure=False,  # True if using HTTPS
    )
except KeyError as e:
    print(f"Configuration error: {e}")
except Exception as e:
    print(f"Remote service setup failed: {e}")

# Handle PostgreSQL connection
try:
    PS = psycopg2.connect(
        database=SETTING["postgres"]["database"],
        user=SETTING["postgres"]["username"],
        password=SETTING["postgres"]["password"],
        host=SETTING["postgres"]["server_url"],
        port=SETTING["postgres"]["port"],
    )
except KeyError as e:
    print(f"PostgreSQL configuration error: {e}")
except Exception as e:
    print(f"Failed to connect to PostgreSQL: {e}")



import ujson
import kerchunk.hdf
from kerchunk.combine import MultiZarrToZarr
import kerchunk.netCDF3
import geopandas as gpd
import os
import shutil
import boto3

from hydroutils.hydro_s3 import boto3_upload_file, boto3_download_file

from hydrodatasource.configs.config import FS
from hydrodatasource.configs.config import RO

so = dict(
    mode="rb", storage_options=RO, default_fill_cache=False, default_cache_type="first"
)  # args to fs.open()
# default_fill_cache=False avoids caching data in between file chunks to lowers memory usage.


class HDFProcessor:
    """
    适用于gpm等Hdf5格式数据。

    Attributes:

    Method:
    """

    def __init__(self):
        pass

    def nc_to_zarr(self, nc_path, json_path):
        with FS.open(nc_path, **so) as infile:
            try:
                h5chunks = kerchunk.hdf.SingleHdf5ToZarr(infile, nc_path)
            except Exception:
                print(nc_path, "未生成！")
                return

            with FS.open(json_path, "wb") as f:
                f.write(ujson.dumps(h5chunks.translate()).encode())

    def multi_to_zarr(
        self, json_paths, file_path, concat_dims=None, identical_dims=None
    ):
        if concat_dims is None:
            concat_dims = ["time"]
        if identical_dims is None:
            identical_dims = ["lat", "lon"]
        json_list = FS.glob(json_paths)
        json_list = [f"s3://{str}" for str in json_list]

        mzz = MultiZarrToZarr(
            json_list,
            target_options=RO,
            remote_protocol="s3",
            remote_options=RO,
            concat_dims=concat_dims,
            identical_dims=identical_dims,
        )

        d = mzz.translate()

        with FS.open(file_path) as f:
            f.write(ujson.dumps(d).encode())


class NC3Processor:
    """
    适用于era5等netCDF3格式数据。

    Attributes:

    Method:
    """

    def __init__(self):
        pass

    def nc_to_zarr(self, nc_path, json_path):
        try:
            h5chunks = kerchunk.netCDF3.netcdf_recording_file(
                f"s3://{nc_path}", storage_options=RO
            )
        except Exception:
            print(nc_path, "未生成！")
            return

        with FS.open(json_path, "wb") as f:
            f.write(ujson.dumps(h5chunks.translate()).encode())

    def multi_to_zarr(
        self, json_paths, file_path, concat_dims=None, identical_dims=None
    ):
        if concat_dims is None:
            concat_dims = ["time"]
        if identical_dims is None:
            identical_dims = ["latitude", "longitude"]
        json_list = FS.glob(json_paths)
        json_list = [f"s3://{str}" for str in json_list]

        mzz = MultiZarrToZarr(
            json_list,
            target_options=RO,
            remote_protocol="s3",
            remote_options=RO,
            concat_dims=concat_dims,
            identical_dims=identical_dims,
        )

        d = mzz.translate()

        with FS.open(file_path) as f:
            f.write(ujson.dumps(d).encode())


def geojson_to_shp(input_geojson, output_folder=None, keep_folder=True):
    """Trans geojson to shp and zip it, return the path of zip file"""
    gdf = gpd.read_file(input_geojson)

    # 根据输入的GeoJSON文件确定输出路径
    if not output_folder:
        output_folder = os.path.join(os.path.dirname(input_geojson), "shp_output")

    # 确保输出文件夹存在
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)

    # 写入Shapefile
    base_name = os.path.splitext(os.path.basename(input_geojson))[0]
    output_shp = os.path.join(output_folder, f"{base_name}.shp")
    gdf.to_file(output_shp)

    # 将Shapefile及其相关文件打包成ZIP文件
    archive_path = shutil.make_archive(
        os.path.join(os.path.dirname(output_folder), base_name), "zip", output_folder
    )

    # 根据参数决定是否删除子文件夹
    if not keep_folder:
        shutil.rmtree(output_folder)

    return archive_path


class GeoProcessor:
    def __init__(self, minio_paras):
        self.boto = boto3.client(
            "s3",
            endpoint_url=minio_paras["endpoint_url"],
            aws_access_key_id=minio_paras["access_key"],
            aws_secret_access_key=minio_paras["secret_key"],
        )
        self.bucket_name = minio_paras["bucket_name"]
        self.endpoint = minio_paras["endpoint_url"].replace("http://", "")

    def upload_geojson(self, gj_local_path, gj_mo_name=None, upload_shp=True):
        """Upload geojson file and optionally its shpfile version."""
        # 如果gj_mo_name没有提供，则使用gj_local_path的文件名
        if gj_mo_name is None:
            gj_mo_name = os.path.basename(gj_local_path)

        if upload_shp:
            shp_zip_path = geojson_to_shp(gj_local_path)
            zip_name = os.path.basename(shp_zip_path)
            boto3_upload_file(self.boto, self.bucket_name, zip_name, shp_zip_path)

    def read_shp(self, shp_name):
        """Read shpfile from minio."""
        return gpd.read_file(
            f"zip+http://{self.endpoint}/{self.bucket_name}/{shp_name}"
        )

    def download_shp(self, shp_name, local_path):
        """Download shpfile from minio."""
        boto3_download_file(self.boto, self.bucket_name, shp_name, local_path)



'''
Author: liutiaxqabs 1498093445@qq.com
Date: 2024-04-24 14:02:00
LastEditors: liutiaxqabs 1498093445@qq.com
LastEditTime: 2024-04-26 10:32:48
FilePath: /hydrodatasource/tests/test_waterlevel_cleaner.py
Description: 这是默认设置,请设置`customMade`, 打开koroFileHeader查看配置 进行设置: https://github.com/OBKoro1/koro1FileHeader/wiki/%E9%85%8D%E7%BD%AE
'''

import pytest
from hydrodatasource.cleaner.waterlevel_cleaner import WaterlevelCleaner
import pandas as pd
import matplotlib.pyplot as plt


def test_anomaly_process():
    # 测试水位数据处理功能
    cleaner = WaterlevelCleaner(
        "/home/liutianxv1/水位sampledatatest.csv", grad_max=0.5, window_size=168
    )
    methods = ["roll", "moving_grad"]
    cleaner.anomaly_process(methods)

    print(cleaner.origin_df)
    print(cleaner.processed_df)

    cleaner.processed_df.to_csv("/home/liutianxv1/水位sampledatatest.csv")



'''
Author: liutiaxqabs 1498093445@qq.com
Date: 2024-04-19 14:00:06
LastEditors: liutiaxqabs 1498093445@qq.com
LastEditTime: 2024-06-21 12:13:04
FilePath: /hydrodatasource/hydrodatasource/cleaner/rainfall_cleaner.py
Description: 这是默认设置,请设置`customMade`, 打开koroFileHeader查看配置 进行设置: https://github.com/OBKoro1/koro1FileHeader/wiki/%E9%85%8D%E7%BD%AE
'''

import numpy as np
import pandas as pd
import xarray as xr
import os
from datetime import datetime, timedelta
import geopandas as gpd
import matplotlib.pyplot as plt
from shapely.geometry import Point, Polygon
from scipy.spatial import Voronoi
import numpy as np
from geopandas.tools import sjoin
from .cleaner import Cleaner


class RainfallCleaner(Cleaner):
    def __init__(
        self,
        data_path,
        era5_path,
        station_file,
        start_time,
        end_time,
        grad_max=200,
        extr_max=200,
        *args,
        **kwargs,
    ):
        self.temporal_list = pd.DataFrame()  # 初始化为空的DataFrame
        self.spatial_list = pd.DataFrame()
        self.grad_max = grad_max
        self.extr_max = extr_max
        self.era5_path = era5_path
        self.station_file = (
            pd.read_csv(station_file, dtype={"STCD": str})
            if isinstance(station_file, str)
            else station_file
        )
        self.start_time, self.end_time = start_time, end_time

        super().__init__(data_path, *args, **kwargs)

    # 数据极大值检验
    def extreme_filter(self, rainfall_data):
        # 创建数据副本以避免修改原始DataFrame
        df = rainfall_data.copy()
        # 设置汛期与非汛期极值阈值
        extreme_value_flood = self.extr_max
        extreme_value_non_flood = self.extr_max / 2
        df["TM"] = pd.to_datetime(df["TM"])
        # 识别汛期
        df["Is_Flood_Season"] = df["TM"].apply(lambda x: 6 <= x.month <= 9)

        # 对超过极值阈值的数据进行处理，将DRP值设置为0
        df.loc[
            (df["Is_Flood_Season"] == True) & (df["DRP"] > extreme_value_flood),
            "DRP",
        ] = 0
        df.loc[
            (df["Is_Flood_Season"] == False) & (df["DRP"] > extreme_value_non_flood),
            "DRP",
        ] = 0

        return df

    # 数据梯度筛查
    def gradient_filter(self, rainfall_data):

        # 原始总降雨量
        original_total_rainfall = rainfall_data["DRP"].sum()

        # 创建数据副本以避免修改原始DataFrame
        df = rainfall_data.copy()

        # 计算降雨量变化梯度
        df["DRP_Change"] = df["DRP"].diff()

        # 汛期与非汛期梯度阈值
        gradient_threshold_flood = self.grad_max
        gradient_threshold_non_flood = self.grad_max / 2

        # 识别汛期
        df["TM"] = pd.to_datetime(df["TM"])
        df["Is_Flood_Season"] = df["TM"].apply(lambda x: 6 <= x.month <= 9)

        # 处理异常值
        df.loc[
            (df["Is_Flood_Season"] == True)
            & (df["DRP_Change"].abs() > gradient_threshold_flood),
            "DRP",
        ] = 0
        df.loc[
            (df["Is_Flood_Season"] == False)
            & (df["DRP_Change"].abs() > gradient_threshold_non_flood),
            "DRP",
        ] = 0

        # 调整后的总降雨量
        adjusted_total_rainfall = df["DRP"].sum()

        # 打印数据总量的变化
        print(f"Original Total Rainfall: {original_total_rainfall} mm")
        print(f"Adjusted Total Rainfall: {adjusted_total_rainfall} mm")
        print(f"Change: {adjusted_total_rainfall - original_total_rainfall} mm")

        # 清理不再需要的列
        df.drop(columns=["DRP_Change", "Is_Flood_Season"], inplace=True)
        return df

    # 数据累计量检验
    def sum_validate_detect(self, rainfall_data):
        """
        检查每个站点每年的总降雨量是否在200到2000毫米之间，并为每个站点生成一个年度降雨汇总表。
        :param rainfall_data: 包含站点代码('STCD')、降雨量('DRP')和时间('TM')的DataFrame
        :return: 新的DataFrame，包含STCD, YEAR, DRP_SUM, IS_REA四列
        """
        # 复制数据并转换日期格式
        df = rainfall_data[
            [
                "STCD",
                "TM",
                "DRP",
            ]
        ].copy()
        df["TM"] = pd.to_datetime(df["TM"])
        df["Year"] = df["TM"].dt.year  # 添加年份列

        # 按站点代码和年份分组，并计算每年的累计降雨量
        grouped = df.groupby(["STCD", "Year"])
        annual_summary = grouped["DRP"].sum().reset_index(name="DRP_SUM")

        # 判断每年的累计降雨量是否在指定范围内
        annual_summary["IS_REA"] = annual_summary["DRP_SUM"].apply(
            lambda x: 200 <= x <= 2000
        )

        return annual_summary

    def era5land_df(self, era5_path, start_time, end_time):
        output_dir = "/ftproot/era5land"
        output_file = os.path.join(output_dir, "tp.csv")

        # 检查是否已存在处理过的文件
        if os.path.exists(output_file):
            print("Using cached data from", output_file)
            return pd.read_csv(output_file)

        # 解析开始和结束时间
        start_date = datetime.strptime(start_time, "%Y-%m-%d")
        end_date = datetime.strptime(end_time, "%Y-%m-%d")

        # 初始化最终的 DataFrame
        final_df = pd.DataFrame()

        # 遍历目录中的所有文件
        for file_name in os.listdir(era5_path):
            if file_name.endswith(".nc"):  # 确保是 NetCDF 文件
                file_path = os.path.join(era5_path, file_name)

                # 打开 NetCDF 数据集
                try:
                    with xr.open_dataset(file_path) as ds:
                        # 提取并四舍五入经纬度数据
                        longitude = np.round(ds["longitude"].values, 1)
                        latitude = np.round(ds["latitude"].values, 1)
                        tp = ds["tp"]

                        # 选择数据集的第一个时间点，通常代表0点
                        tp_at_first_time = tp.isel(time=0)

                        # 获取时间信息
                        date_str = str(tp_at_first_time["time"].values)[:10]
                        data_date = datetime.strptime(date_str, "%Y-%m-%d")

                        # 检查是否在指定的时间范围内
                        if start_date <= data_date <= end_date:
                            tp_value = tp_at_first_time.values.flatten()

                            # 创建站点ID
                            station_ids = [
                                "era5land_{:.1f}_{:.1f}".format(lon, lat)
                                for lon in longitude
                                for lat in latitude
                            ]

                            # 创建临时 DataFrame
                            temp_df = pd.DataFrame(
                                {
                                    "ID": station_ids,
                                    "LON": np.repeat(longitude, len(latitude)),
                                    "LAT": np.tile(latitude, len(longitude)),
                                    "TP": tp_value,
                                    "TM": (data_date - timedelta(days=1)).strftime(
                                        "%Y-%m-%d"
                                    ),  # 使用前一天的日期
                                }
                            )

                            # 将临时 DataFrame 添加到最终 DataFrame
                            final_df = pd.concat([final_df, temp_df], ignore_index=True)
                except Exception as e:
                    print(f"Failed to process file {file_name}: {e}")

        # 空数据检查
        if final_df.empty:
            print("No data processed. Please check the input files and date range.")
            return None

        # 转换 TM 列为 datetime 类型，以便提取年份
        final_df["TM"] = pd.to_datetime(final_df["TM"])

        # 创建一个新的列 'Year' 来存储年份
        final_df["Year"] = final_df["TM"].dt.year

        # 按 ID 和 Year 分组，计算每个站点每年的 TP 总和
        annual_sum_df = (
            final_df.groupby(["ID", "Year"])
            .agg(
                {
                    "LON": "first",  # 取第一个经度作为代表
                    "LAT": "first",  # 取第一个纬度作为代表
                    "TP": "sum",  # 求和降水量
                }
            )
            .reset_index()
        )
        annual_sum_df.to_csv("/ftproot/era5land/tp.csv", index=False)
        return annual_sum_df

    # 空间信息筛选雨量站（ERA5-LAND校准）
    def spatial_era5land_detect(self, rainfall_data):
        # 截获起止时间计算era5land数据
        era5land_df = self.era5land_df(
            era5_path=self.era5_path, start_time=self.start_time, end_time=self.end_time
        )
        rainfall_df = self.sum_validate_detect(rainfall_data=rainfall_data)
        # 拿站点经纬度找最佳匹配得站点
        rainfall = pd.merge(rainfall_df, self.station_file, on="STCD", how="left")

        # 添加新列以存储匹配的ERA5 TP值
        rainfall["ERA5_TP"] = np.nan

        for index, rain_row in rainfall.iterrows():
            # 在ERA5数据中匹配网格
            matched = era5land_df[
                (era5land_df["LON"] <= rain_row["LON"])
                & (era5land_df["LON"] + 0.1 > rain_row["LON"])
                & (era5land_df["LAT"] - 0.1 < rain_row["LAT"])
                & (era5land_df["LAT"] >= rain_row["LAT"])
                & (era5land_df["Year"] == rain_row["Year"])
            ]

            if not matched.empty:
                # 如果找到匹配，取第一条匹配记录的TP值
                rainfall.at[index, "ERA5_TP"] = matched.iloc[0]["TP"]
            else:
                # 如果没有找到匹配，设置ERA5 TP值为NaN
                rainfall.at[index, "ERA5_TP"] = np.nan

        # 判断合理性
        rainfall["IS_REA"] = False

        # 判断条件并设置 IS_REA 列的值
        valid_indices = (
            (rainfall["ERA5_TP"].notnull())  # ERA5_TP 不为空
            & (rainfall["DRP_SUM"].notnull())  # DRP_SUM 不为空
            & (
                0.8 * rainfall["ERA5_TP"] <= rainfall["DRP_SUM"]
            )  # DRP_SUM 大于等于 0.8 * ERA5_TP
            & (
                rainfall["DRP_SUM"] <= 1.2 * rainfall["ERA5_TP"]
            )  # DRP_SUM 小于等于 1.2 * ERA5_TP
        )
        rainfall.loc[valid_indices, "IS_REA"] = True

        return rainfall[["STCD", "Year", "DRP_SUM", "LON", "LAT", "ERA5_TP", "IS_REA"]]

    def anomaly_process(self, methods=None):
        super().anomaly_process(methods)
        rainfall_data = self.origin_df
        for method in methods:
            if method == "extreme":
                rainfall_data = self.extreme_filter(rainfall_data=rainfall_data)
            elif method == "gradient":
                rainfall_data = self.gradient_filter(rainfall_data=rainfall_data)
            elif method == "detect_sum":
                self.temporal_list = self.sum_validate_detect(
                    rainfall_data=rainfall_data
                )
            elif method == "detect_era5":
                self.spatial_list = self.spatial_era5land_detect(
                    rainfall_data=rainfall_data
                )
            else:
                print("please check your method name")

        # self.processed_df["DRP"] = rainfall_data["DRP"] # 最终结果赋值给processed_df
        # 新增一列进行存储
        self.processed_df[methods[0]] = rainfall_data["DRP"]


class RainfallAnalyzer:
    def __init__(
        self,
        stations_csv_path=None,
        shp_folder=None,
        rainfall_data_folder=None,
        output_folder=None,
        output_log=None,
        output_plot=None,
        lower_bound=None,
        upper_bound=None,
    ):
        self.stations_csv_path = stations_csv_path
        self.shp_folder = shp_folder
        self.rainfall_data_folder = rainfall_data_folder
        self.output_folder = output_folder
        self.output_log = output_log
        self.output_plot = output_plot
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def filter_and_save_csv(self):
        """
        筛选降雨数据，根据每年的总降雨量（DRP）进行过滤，保留符合最低和最高阈值的数据。

        参数：
        input_folder - 包含降雨数据的文件夹路径。
        lower_bound - 降雨量最低阈值。
        upper_bound - 降雨量最高阈值。

        返回：
        过滤后的降雨数据DataFrame。
        """
        print("Filtering data by yearly total DRP")
        input_folder = self.rainfall_data_folder
        filtered_data_list = []
        for file in os.listdir(input_folder):
            if file.endswith(".csv"):
                file_path = os.path.join(input_folder, file)
                data = pd.read_csv(file_path)
                data["TM"] = pd.to_datetime(data["TM"], errors='coerce')
                #data["TM"] = pd.to_datetime(data["TM"], format="%Y-%m-%d %H:%M:%S.%f", errors='coerce')
                data["DRP"] = data["DRP"].astype(float)

                data["ID"] = file.replace(".csv", "")
                for year, group in data.groupby(data["TM"].dt.year):
                    drp_sum = group["DRP"].sum()
                    if self.lower_bound <= drp_sum <= self.upper_bound:
                        # print(f"File {file} contains valid data for year {year} with DRP sum {drp_sum}")
                        filtered_data_list.append(group)
        if filtered_data_list:
            return pd.concat(filtered_data_list, ignore_index=True)
        else:
            return pd.DataFrame()

    def read_data(self, basin_shp_path):
        """
        读取站点信息和流域shapefile数据。

        参数：
        stations_csv_path - 站点信息CSV文件路径。
        basin_shp_path - 流域shapefile文件路径。

        返回：
        stations_df - 站点信息DataFrame。
        basin - 流域shapefile的GeoDataFrame。
        """
        stations_df = pd.read_csv(self.stations_csv_path)
        stations_df.dropna(subset=["LON", "LAT"], inplace=True)
        basin = gpd.read_file(basin_shp_path)
        return stations_df, basin

    def process_stations(self, stations_df, basin):
        """
        筛选位于流域内部的站点数据。

        参数：
        stations_df - 站点信息DataFrame。
        basin - 流域shapefile的GeoDataFrame。

        返回：
        stations_within_basin - 位于流域内部的站点GeoDataFrame。
        """
        print("Processing stations within the basin")
        gdf_stations = gpd.GeoDataFrame(
            stations_df,
            geometry=[Point(xy) for xy in zip(stations_df.LON, stations_df.LAT)],
            crs="EPSG:4326",
        )
        gdf_stations = gdf_stations.to_crs(basin.crs)
        stations_within_basin = sjoin(gdf_stations, basin, predicate="within")
        print(f"Found {len(stations_within_basin)} stations within the basin")
        return stations_within_basin

    def calculate_voronoi_polygons(self, stations, basin):
        """
        计算泰森多边形并裁剪至流域边界。

        参数：
        stations - 位于流域内部的站点GeoDataFrame。
        basin - 流域shapefile的GeoDataFrame。

        返回：
        clipped_polygons - 裁剪后的泰森多边形GeoDataFrame。
        """
        if len(stations) < 2:
            stations["original_area"] = np.nan
            stations["clipped_area"] = np.nan
            stations["area_ratio"] = 1.0
            return stations

        # 获取流域边界的最小和最大坐标，构建边界框
        x_min, y_min, x_max, y_max = basin.total_bounds

        # 扩展边界框
        x_min -= 1.0 * (x_max - x_min)
        x_max += 1.0 * (x_max - x_min)
        y_min -= 1.0 * (y_max - y_min)
        y_max += 1.0 * (y_max - y_min)

        bounding_box = np.array(
            [[x_min, y_min], [x_max, y_min], [x_max, y_max], [x_min, y_max]]
        )

        # 提取站点坐标
        points = np.array([point.coords[0] for point in stations.geometry])

        # 将站点坐标与边界框点结合，确保Voronoi多边形覆盖整个流域
        points_extended = np.concatenate((points, bounding_box), axis=0)

        # 计算Voronoi图
        vor = Voronoi(points_extended)

        # 提取每个点对应的Voronoi区域
        regions = [vor.regions[vor.point_region[i]] for i in range(len(points))]

        # 生成多边形
        polygons = [
            Polygon([vor.vertices[i] for i in region if i != -1])
            for region in regions
            if -1 not in region
        ]

        # 创建GeoDataFrame
        gdf_polygons = gpd.GeoDataFrame(geometry=polygons, crs=stations.crs)
        gdf_polygons["STCD"] = stations["STCD"].values
        gdf_polygons["original_area"] = gdf_polygons.geometry.area

        # 计算流域的总面积
        basin_area = basin.geometry.area.sum()
        print(f"Basin area: {basin_area}")

        # 计算原始泰森多边形的总面积
        total_original_area = gdf_polygons["original_area"].sum()
        print(f"Total original Voronoi polygons area: {total_original_area}")

        # 将多边形裁剪到流域边界
        clipped_polygons = gpd.clip(gdf_polygons, basin)
        clipped_polygons["clipped_area"] = clipped_polygons.geometry.area
        clipped_polygons["area_ratio"] = (
            clipped_polygons["clipped_area"] / clipped_polygons["clipped_area"].sum()
        )

        # 计算裁剪后泰森多边形的总面积
        total_clipped_area = clipped_polygons["clipped_area"].sum()
        print(f"Total clipped Voronoi polygons area: {total_clipped_area}")

        # 打印年度数据汇总并将其追加到日志文件中
        log_file = self.output_log
        with open(log_file, "a") as f:
            log_entries = [
                f"Basin area: {basin_area}",
                f"Total original Voronoi polygons area: {total_original_area}",
                f"Total clipped Voronoi polygons area: {total_clipped_area}",
            ]
            for entry in log_entries:
                print(entry)
                f.write(entry + "\n")

        return clipped_polygons

    def calculate_weighted_rainfall(self, thiesen_polygons, rainfall_df):
        """
        计算加权平均降雨量。

        参数：
        thiesen_polygons - 泰森多边形GeoDataFrame。
        rainfall_df - 降雨数据DataFrame。

        返回：
        weighted_average_rainfall - 加权平均降雨量DataFrame。
        """
        thiesen_polygons["STCD"] = thiesen_polygons["STCD"].astype(str)
        rainfall_df["STCD"] = rainfall_df["STCD"].astype(str)

        # 合并泰森多边形和降雨数据
        merged_data = pd.merge(thiesen_polygons, rainfall_df, on="STCD")

        # 计算加权降雨量
        merged_data["weighted_rainfall"] = (
            merged_data["DRP"] * merged_data["area_ratio"]
        )

        # 按时间分组并计算加权平均降雨量
        weighted_average_rainfall = (
            merged_data.groupby("TM")["weighted_rainfall"].sum().reset_index()
        )

        return weighted_average_rainfall

    def display_results(
        self,
        year,
        valid_stations,
        thiesen_polygons_year,
        yearly_data,
        average_rainfall,
        basin,
    ):
        """
        显示处理结果，包括地图展示、站点信息、降雨量信息和平均降雨量。

        参数：
        year - 当前处理的年份。
        valid_stations - 符合条件的站点GeoDataFrame。
        yearly_data - 当前年份的降雨数据DataFrame。
        average_rainfall - 加权平均降雨量DataFrame。
        basin - 流域shapefile的GeoDataFrame。
        """
        print(f"Displaying results for year {year}")

        # 绘制经纬度图像
        fig, ax = plt.subplots(figsize=(10, 10))
        basin.plot(ax=ax, color="lightgrey", edgecolor="black")
        thiesen_polygons_year.plot(
            ax=ax, facecolor="blue", edgecolor="black", markersize=50
        )
        valid_stations.plot(ax=ax, color="red", markersize=50)
        plt.title(f"Stations within basin {basin['BASIN_ID'].iloc[0]} for year {year}")
        plt.xlabel("Longitude")
        plt.ylabel("Latitude")
        # 生成文件名
        file_name = f"{basin['BASIN_ID'].iloc[0]}_{year}.png"
        file_path = f"{self.output_plot}/{file_name}"
        # 保存图像
        plt.savefig(file_path)
        # plt.show()

        # 输出站点名称和数量
        station_names = valid_stations["ID"].tolist()
        station_count = len(station_names)
        print(f"Stations for year {year}: {station_names}")
        print(f"Total number of stations: {station_count}")

        # 输出对应年份的数据
        filtered_yearly_data = yearly_data[yearly_data["ID"].isin(station_names)]

        yearly_summary = (
            filtered_yearly_data.groupby("ID")
            .agg({"STCD": "first", "DRP": "sum"})
            .reset_index()
        )
        print(f"Yearly data summary for year {year}:\n{yearly_summary}")

        # 输出平均雨量数据
        mean_rainfall = average_rainfall["mean_rainfall"].sum()
        print(f"Average rainfall for year {year}: {mean_rainfall}")

        # 追加日志
        # 打印年度数据汇总并将其追加到日志文件中
        log_file = self.output_log
        with open(log_file, "a") as f:
            log_entries = [
                f"BASINS: {basin['BASIN_ID'].iloc[0]}",
                f"Displaying results for year {year}",
                f"Stations for year {year}: {station_names}",
                f"Total number of stations: {station_count}",
                f"Yearly data summary for year {year}:\n{yearly_summary}",
                f"Average rainfall for year {year}: {mean_rainfall}\n",
            ]
            for entry in log_entries:
                print(entry)
                f.write(entry + "\n")

    def process_basin(self, basin_shp_path, filtered_data):
        """
        处理每个流域的降雨数据，计算泰森多边形和面平均降雨量。

        参数：
        basin_shp_path - 流域shapefile文件路径。
        stations_csv_path - 站点信息CSV文件路径。
        filtered_data - 预先过滤的降雨数据DataFrame。
        output_folder - 输出文件夹路径。
        """
        all_years_rainfall = []
        stations_df, basin = self.read_data(basin_shp_path)

        years = filtered_data["TM"].dt.year.unique()

        for year in sorted(years):
            print(
                f"Processing basin {os.path.basename(basin_shp_path)} for year {year}"
            )
            # 打印年度数据汇总并将其追加到日志文件中
            log_file = self.output_log
            with open(log_file, "a") as f:
                f.write(
                    f"Processing basin {os.path.basename(basin_shp_path)} for year {year}\n"
                )
            yearly_data = filtered_data[filtered_data["TM"].dt.year == year]

            if yearly_data.empty:
                print(
                    f"No valid data for basin {os.path.basename(basin_shp_path)} in year {year}"
                )
                # 打印年度数据汇总并将其追加到日志文件中
                log_file = self.output_log
                with open(log_file, "a") as f:
                    f.write(
                        f"No valid stations for basin {os.path.basename(basin_shp_path)} in year {year}\n"
                    )
                continue

            # 筛选符合条件的每年站点数据
            yearly_stations = yearly_data["ID"].unique()
            valid_stations = self.process_stations(stations_df, basin)
            valid_stations = valid_stations[valid_stations["ID"].isin(yearly_stations)]

            if valid_stations.empty:
                print(
                    f"No valid stations for basin {os.path.basename(basin_shp_path)} in year {year}"
                )
                # 打印年度数据汇总并将其追加到日志文件中
                log_file = self.output_log
                with open(log_file, "a") as f:
                    f.write(
                        f"No valid stations for basin {os.path.basename(basin_shp_path)} in year {year}\n"
                    )

                continue

            thiesen_polygons_year = self.calculate_voronoi_polygons(
                valid_stations, basin
            )
            average_rainfall = self.calculate_weighted_rainfall(
                thiesen_polygons_year, yearly_data
            )
            average_rainfall.columns = ["TM", "mean_rainfall"]
            basin_id = os.path.splitext(os.path.basename(basin_shp_path))[0]
            average_rainfall["ID"] = basin_id
            all_years_rainfall.append(average_rainfall)

            # 调用展示函数
            self.display_results(
                year,
                valid_stations,
                thiesen_polygons_year,
                yearly_data,
                average_rainfall,
                basin,
            )

        if all_years_rainfall:
            final_result = pd.concat(all_years_rainfall, ignore_index=True)
            basin_output_folder = os.path.join(self.output_folder, basin_id)
            os.makedirs(basin_output_folder, exist_ok=True)
            output_file = os.path.join(basin_output_folder, f"{basin_id}_rainfall.csv")
            final_result.to_csv(output_file, index=False)
            print(f"Result for basin {basin_id} saved to {output_file}")
        else:
            print(
                f"No valid data for basin {os.path.splitext(os.path.basename(basin_shp_path))[0]}"
            )

    def basins_polygon_mean(self):
        """
        主函数，执行整个数据处理流程。

        参数：
        stations_csv_path - 站点信息CSV文件路径。
        shp_folder - 流域shapefile文件夹路径。
        rainfall_data_folder - 降雨数据文件夹路径。
        lower_bound - 降雨量最低阈值。
        upper_bound - 降雨量最高阈值。
        output_folder - 输出文件夹路径。
        """
        # 先筛选降雨数据，保留符合最低阈值的数据
        filtered_data = self.filter_and_save_csv()
        for shp_file in os.listdir(self.shp_folder):
            if shp_file.endswith(".shp"):
                basin_shp_path = os.path.join(self.shp_folder, shp_file)
                self.process_basin(basin_shp_path, filtered_data)

    # 添加时间一致性检验功能
    def check_time_consistency(self, df, hours=24):
        # 假设时间列为'TM'和降雨量列为'DRP'
        datetime_col = "TM"
        rainfall_col = "DRP"

        # 解析日期时间列
        df[datetime_col] = pd.to_datetime(df[datetime_col])

        # 按时间排序
        df = df.sort_values(by=datetime_col)

        # 添加一个布尔列来标记异常
        df["is_anomaly"] = False

        # 检查是否有连续24小时降雨量完全一致且非零的情况
        for i in range(len(df) - hours + 1):
            window = df.iloc[i : i + hours]
            # 过滤掉NaN值
            if window[rainfall_col].isna().sum() == 0:
                if (
                    len(window[rainfall_col].unique()) == 1
                    and window[rainfall_col].iloc[0] != 0
                ):
                    df.loc[window.index, "is_anomaly"] = True

        return df

    def time_consistency(self):
        all_anomalies = pd.DataFrame()

        # 遍历rainfall_data_folder中的所有文件
        for file_name in os.listdir(self.rainfall_data_folder):
            file_path = os.path.join(self.rainfall_data_folder, file_name)

            if os.path.isfile(file_path) and file_name.endswith(".csv"):
                # 读取CSV文件
                df = pd.read_csv(file_path)

                # 检查时间一致性
                df = self.check_time_consistency(df)

                # 提取标记为异常的记录
                anomalies = df[df["is_anomaly"]]

                # 将异常数据添加到总的DataFrame中
                all_anomalies = pd.concat([all_anomalies, anomalies], ignore_index=True)

        # 将所有异常数据保存到一个txt文件中
        output_file = os.path.join(self.output_folder, "time_consistency_anomalies.txt")
        with open(output_file, "w") as f:
            f.write(all_anomalies.to_string(index=False))

        print(f"异常数据已保存到 {output_file}")

    # 添加空间一致性检验功能
    def spatial_consistency(self):
        """
        主函数，执行整个数据处理流程。

        参数：
        stations_csv_path - 站点信息CSV文件路径。
        shp_folder - 流域shapefile文件夹路径。
        rainfall_data_folder - 降雨数据文件夹路径。
        output_folder - 输出文件夹路径。
        """
        self.lower_bound = -1
        self.upper_bound = 50000
        filtered_data = self.filter_and_save_csv()
        for shp_file in os.listdir(self.shp_folder):
            if shp_file.endswith(".shp"):
                basin_shp_path = os.path.join(self.shp_folder, shp_file)
                #########################################################
                # 以下是空间一致性检验的代码
                #########################################################



'''
Author: liutiaxqabs 1498093445@qq.com
Date: 2024-04-19 14:00:16
LastEditors: liutiaxqabs 1498093445@qq.com
LastEditTime: 2024-08-05 11:50:28
FilePath: /hydrodatasource/hydrodatasource/cleaner/streamflow_cleaner.py
Description: 这是默认设置,请设置`customMade`, 打开koroFileHeader查看配置 进行设置: https://github.com/OBKoro1/koro1FileHeader/wiki/%E9%85%8D%E7%BD%AE
'''

from .cleaner import Cleaner
import xarray as xr
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from scipy.signal import cwt, morlet, butter, filtfilt
from scipy.fft import fft, ifft, fftfreq
from scipy.optimize import curve_fit
import os
from tqdm import tqdm


class StreamflowCleaner(Cleaner):
    def __init__(
        self,
        data_path,
        window_size=14,
        stride=1,
        cutoff_frequency=0.035,
        time_step=1.0,
        iterations=3,
        sampling_rate=1.0,
        order=5,
        cwt_row=2,
        *args,
        **kwargs,
    ):
        self.window_size = window_size
        self.stride = stride
        self.cutoff_frequency = cutoff_frequency
        self.time_step = time_step
        self.iterations = iterations
        self.sampling_rate = sampling_rate
        self.order = order
        self.cwt_row = cwt_row
        super().__init__(data_path, *args, **kwargs)

    def data_balanced(self, origin_data, transform_data):
        """
        对一维流量数据进行总量平衡变换。
        :origin_data: 原始一维流量数据。
        :transform_data: 平滑转换后的一维流量数据。
        """
        # Calculate the flow balance factor and keep the total volume consistent
        streamflow_data_before = np.sum(origin_data)
        streamflow_data_after = np.sum(transform_data)
        scaling_factor = streamflow_data_before / streamflow_data_after
        balanced_data = transform_data * scaling_factor

        print(f"Total flow (before smoothing): {streamflow_data_before}")
        print(f"Total flow (after smoothing): {np.sum(balanced_data)}")
        return balanced_data

    def moving_average(self, streamflow_data):
        """
        对流量数据应用滑动平均进行平滑处理，并保持流量总量平衡。
        :param streamflow_data: 输入的流量数据数组
        :return: 平滑处理后的流量数据
        """
        # 将流量数据转换为 pandas Series
        streamflow_series = streamflow_data

        # 应用中心滑动平均
        smoothed_series = streamflow_series.rolling(
            window=self.window_size, center=True
        ).mean()

        # 填充由于滚动窗口导致的起始和结束的 NaN 值
        smoothed_series.bfill(inplace=True)  # 用后面的值填充前面的 NaN
        smoothed_series.ffill(inplace=True)  # 用前面的值填充后面的 NaN

        # 将平滑数据中的负值置为0
        smoothed_series[smoothed_series < 0] = 0

        # 将结果转换回 numpy 数组
        smoothed_data = smoothed_series

        return self.data_balanced(streamflow_data, smoothed_data)

    def kalman_filter(self, streamflow_data):
        """
        对流量数据应用卡尔曼滤波进行平滑处理，并保持流量总量平衡。
        :param streamflow_data: 原始流量数据
        """
        A = np.array([[1]])
        H = np.array([[1]])
        Q = np.array([[0.01]])
        R = np.array([[0.01]])
        X_estimated = np.array([streamflow_data[0]])
        P_estimated = np.eye(1) * 0.01
        estimated_states = []

        for measurement in streamflow_data:
            # predict
            X_predicted = A.dot(X_estimated)
            P_predicted = A.dot(P_estimated).dot(A.T) + Q

            # update
            measurement_residual = measurement - H.dot(X_predicted)
            S = H.dot(P_predicted).dot(H.T) + R
            K = P_predicted.dot(H.T).dot(np.linalg.inv(S))  # kalman gain
            X_estimated = X_predicted + K.dot(measurement_residual)
            P_estimated = P_predicted - K.dot(H).dot(P_predicted)
            estimated_states.append(X_estimated.item())

        estimated_states = np.array(estimated_states)

        # Apply non-negative constraints
        estimated_states[estimated_states < 0] = 0
        return self.data_balanced(streamflow_data, estimated_states)

    def adjust_window(self, window):
        if window.count() == 0:
            return np.nan  # 如果窗口内全是NaN，则返回NaN
        adjusted_window = window.copy()
        return adjusted_window.mean()  # 返回窗口的平均值或其他适当的聚合值

    def rolling_with_stride(self, df, func):
        # 初始化与原始 DataFrame 长度相同的 NaN 序列
        results = pd.Series(np.nan, index=df.index)
        # 遍历数据，步长为stride
        for i in range(0, len(df) - self.window_size + 1, self.stride):
            window = df[i : i + self.window_size]
            result = func(window)
            # 计算窗口中心的索引
            center_index = i + self.window_size // 2
            # 仅在中心索引处填充结果
            results.iloc[center_index] = result

        return results

    def moving_average_difference(self, streamflow_data):
        """
        对流量数据应用滑动平均差算法进行平滑处理，并保持流量总量平衡。
        :window_size: 滑动窗口的大小
        """
        streamflow_data_series = pd.Series(streamflow_data)
        # Calculate the forward moving average（MU）
        forward_ma = streamflow_data_series.rolling(
            window=self.window_size, min_periods=1
        ).mean()

        # Calculate the backward moving average（MD）
        backward_ma = (
            streamflow_data_series.iloc[::-1]
            .rolling(window=self.window_size, min_periods=1)
            .mean()
            .iloc[::-1]
        )

        # Calculate the difference between the forward and backward sliding averages
        ma_difference = abs(forward_ma - backward_ma)

        # Apply non-negative constraints
        ma_difference[ma_difference < 0] = 0
        return self.data_balanced(streamflow_data, ma_difference.to_numpy())

    def quadratic_function(self, x, a, b, c):
        return a * x**2 + b * x + c

    def robust_fitting(self, streamflow_data, k=1.5):
        """
        对流量数据应用抗差修正算法进行平滑处理，并保持流量总量平衡。
        默认采用二次曲线进行拟合优化，该算法处理性能较差
        """
        time_steps = np.arange(len(streamflow_data))
        params, _ = curve_fit(self.quadratic_function, time_steps, streamflow_data)
        smoothed_streamflow = self.quadratic_function(time_steps, *params)
        residuals = streamflow_data - smoothed_streamflow
        m = len(streamflow_data)
        sigma = np.sqrt(np.sum(residuals**2) / (m - 1))

        for _ in range(10):
            weights = np.where(
                np.abs(residuals) <= k * sigma, 1, k * sigma / np.abs(residuals)
            )
            sigma = np.sqrt(np.sum(weights * residuals**2) / (m - 1))

        corrected_streamflow = (
            weights * streamflow_data + (1 - weights) * smoothed_streamflow
        )
        corrected_streamflow[corrected_streamflow < 0] = 0
        return self.data_balanced(streamflow_data, corrected_streamflow)

    def lowpass_filter(self, streamflow_data):
        """
        对一维流量数据应用调整后的低通滤波器。
        :cutoff_frequency: 低通滤波器的截止频率。
        :sampling_rate: 数据的采样率。
        :order: 滤波器的阶数，默认为5。
        """

        def apply_low_pass_filter(signal, cutoff_frequency, sampling_rate, order=5):
            nyquist_frequency = 0.5 * sampling_rate
            normalized_cutoff = cutoff_frequency / nyquist_frequency
            b, a = butter(order, normalized_cutoff, btype="low", analog=False)
            filtered_signal = filtfilt(b, a, signal)
            return filtered_signal

        # Apply a low-pass filter
        low_pass_filtered_signal = apply_low_pass_filter(
            streamflow_data, self.cutoff_frequency, self.sampling_rate, self.order
        )

        # Apply non-negative constraints
        low_pass_filtered_signal[low_pass_filtered_signal < 0] = 0

        return self.data_balanced(streamflow_data, low_pass_filtered_signal)

    def FFT(self, streamflow_data):
        """
        对流量数据进行迭代的傅里叶滤波处理，包括非负值调整和流量总量调整。
        :cutoff_frequency: 傅里叶滤波的截止频率。
        :time_step: 数据采样间隔。
        :iterations: 迭代次数。
        """
        current_signal = streamflow_data.to_numpy().copy()

        for _ in range(self.iterations):
            n = len(current_signal)
            yf = fft(current_signal)
            xf = fftfreq(n, d=self.time_step)

            # Applied frequency filtering
            yf[np.abs(xf) > self.cutoff_frequency] = 0

            # FFT and take the real part
            filtered_signal = ifft(yf).real

            # Apply non-negative constraints
            filtered_signal[filtered_signal < 0] = 0

            # Adjust the total flow to match the original flow
            current_signal = self.data_balanced(streamflow_data, filtered_signal)

        return current_signal

    def wavelet(self, streamflow_data):
        """
        对一维流量数据进行小波变换分析前后拓展数据以减少边缘失真，然后调整总流量。
        :cwt_row: 小波变换中使用的特定宽度。
        """
        streamflow_data_array = streamflow_data.to_numpy().copy()
        # Expand the data edge by 24 lines on each side
        extended_data = np.concatenate(
            [
                np.full(
                    24, streamflow_data_array[0]
                ),  # Expand the first 24 lines with the first element
                streamflow_data,
                np.full(
                    24, streamflow_data_array[-1]
                ),  # Expand the last 24 lines with the last element
            ]
        )
        widths = np.arange(1, 31)
        # Wavelet transform by Morlet wavelet directly
        extended_cwt = cwt(extended_data, morlet, widths)
        scaled_cwtmatr = np.abs(extended_cwt)

        # Select a specific width for analysis (can be briefly understood as selecting a cutoff frequency)
        cwt_row_extended = scaled_cwtmatr[self.cwt_row, :]

        # Remove the extended part
        adjusted_cwt_row = cwt_row_extended[24:-24]
        adjusted_cwt_row[adjusted_cwt_row < 0] = 0
        return self.data_balanced(streamflow_data, adjusted_cwt_row)

    def adaptive_moving_average(
        self,
        streamflow_data,
        threshold=100,
        initial_window=168,
        min_window=24,
        max_window=360,
        decay_factor=2,
    ):
        # 确保输入是 pandas Series
        if not isinstance(streamflow_data, pd.Series):
            raise ValueError("输入的数据必须是 pandas Series")

        # 创建一个与原始数据长度相同的Series
        smoothed_data = pd.Series(index=streamflow_data.index, dtype=float)
        current_window = initial_window

        for i, date in enumerate(streamflow_data.index):
            # 获取当前处理节点的值
            current_value = streamflow_data[date]

            # 调整窗口大小
            if current_value >= threshold:
                current_window = max(min_window, current_window // decay_factor)
            else:
                current_window = min(max_window, current_window * decay_factor)

            half_window = current_window // 2

            # 计算窗口的起始和结束时间，处理边界情况
            if i < half_window:
                start_date = streamflow_data.index[0]
            else:
                start_date = date - pd.DateOffset(hours=half_window)

            if i + half_window >= len(streamflow_data):
                end_date = streamflow_data.index[-1]
            else:
                end_date = date + pd.DateOffset(hours=half_window)

            # 计算窗口内的平均值
            window_data = streamflow_data[start_date:end_date]
            smoothed_value = window_data.mean()
            smoothed_data.loc[date] = smoothed_value

        return smoothed_data

    # 使用中心滑动平均处理洪水期间数据
    def update_flood_periods_with_moving_average(
        self, combined_df, flow_division, window_size=1, columns=None
    ):
        for _, row in flow_division.iterrows():
            start_time = row["BEGINNING_FLOW"]
            end_time = row["END_FLOW"]
            mask = (combined_df.index >= start_time) & (combined_df.index <= end_time)
            combined_df.loc[mask, columns] = self.moving_average(
                combined_df.loc[mask, "INQ"], window_size
            )
        return combined_df

    def EMA(self, streamflow_data):
        # 访问时间序列
        df = self.origin_df.copy()

        # streamflow_data数据是插补过的
        df["INQQ"] = np.nan
        df["INQQ"] = streamflow_data

        # 将 'TM' 列转换为日期时间格式
        df["TM"] = pd.to_datetime(df["TM"], errors="coerce")

        # 设置 'TM' 列为索引
        df.set_index("TM", inplace=True)

        # 去重索引，保留最后一个
        df = df[~df.index.duplicated(keep="last")]

        # 分段处理
        # 计算不同窗口的滑动平均
        df["INQA"] = self.adaptive_moving_average(
            df["INQQ"], threshold=200, initial_window=56, min_window=4, max_window=56
        )
        df["INQB"] = self.adaptive_moving_average(
            df["INQQ"], threshold=200, initial_window=140, min_window=28, max_window=140
        )

        # 创建新的INQQ列，根据月份替换数据
        df["INQC"] = np.where(
            df.index.month.isin([5, 6, 7, 8, 9, 10]), df["INQA"], df["INQB"]
        )

        # 处理场次洪水部分
        # flow_division_path = 'biliu_flow_division.csv'  # 洪水场次数据文件路径
        # flow_division = pd.read_csv(flow_division_path)
        # flow_division['BEGINNING_FLOW'] = pd.to_datetime(flow_division['BEGINNING_FLOW'])
        # flow_division['END_FLOW'] = pd.to_datetime(flow_division['END_FLOW'])

        # 更新洪水期间的 INQ 数据
        # df['INQD'] =df['INQC']
        # df = self.update_flood_periods_with_moving_average(df, flow_division,columns = 'INQC', window_size=1)
        # 合并滑动平均结果到 EMA 列
        df["EMA"] = df["INQC"]

        # 进行总量平衡
        df["EMA"] = self.data_balanced(streamflow_data, df["EMA"])

        return df["EMA"]

    def anomaly_process(self, methods=None):
        super().anomaly_process(methods)
        self.origin_df["INQ"] = pd.to_numeric(self.origin_df["INQ"], errors="coerce")
        self.origin_df["TM"] = pd.to_datetime(self.origin_df["TM"], errors="coerce")
        streamflow_data = self.origin_df["INQ"].copy()
        # 使用插值填充缺失值
        streamflow_data = streamflow_data.interpolate().fillna(0)

        for method in methods:
            if method == "moving_average":
                streamflow_data = self.moving_average(streamflow_data=streamflow_data)
            elif method == "kalman":
                streamflow_data = self.kalman_filter(streamflow_data=streamflow_data)
            elif method == "moving_average_diff":
                streamflow_data = self.moving_average_difference(
                    streamflow_data=streamflow_data
                )
            elif method == "robfit":
                streamflow_data = self.robust_fitting(streamflow_data=streamflow_data)
            elif method == "lowpass":
                streamflow_data = self.lowpass_filter(streamflow_data=streamflow_data)
            elif method == "FFT":
                streamflow_data = self.FFT(streamflow_data=streamflow_data)
            elif method == "wavelet":
                streamflow_data = self.wavelet(streamflow_data=streamflow_data)
            elif method == "rolling_mean":
                streamflow_data = self.rolling_with_stride(
                    df=streamflow_data, func=self.adjust_window
                )
                # 确保索引一致
                streamflow_data.index = self.origin_df["INQ"].index
                streamflow_data.fillna(self.origin_df["INQ"], inplace=True)
            elif method == "EMA":
                streamflow_data = self.EMA(streamflow_data=streamflow_data)
                streamflow_data.index = self.origin_df["INQ"].index

            else:
                print("please check your method name")

        # 新增一列进行存储
        self.processed_df[methods[0]] = streamflow_data

        # 去除提前插补的缺失值
        self.processed_df[methods[0]][self.origin_df["INQ"].isna()] = np.nan


class StreamflowBacktrack:
    def __init__(self, data_folder, output_folder,file_name = None):
        self.data_folder = data_folder
        self.output_folder = data_folder
        self.file_name = file_name

    def clean_W(self, file_path, output_folder):
        data = pd.read_csv(file_path)
        # 计算与前一行的差异
        data["diff_prev"] = abs(data["W"] - data["W"].shift(1))

        # 计算与后一行的差异
        data["diff_next"] = abs(data["W"] - data["W"].shift(-1))

        # 标记需要设置为 NaN 的行
        data["set_nan"] = (data["diff_prev"] > 200) | (data["diff_next"] > 200)

        # 如果与前一行或后一行的差异超过200，则设置为 NaN
        data.loc[data["set_nan"], "W"] = np.nan

        # 输出被设置为 NaN 的行
        print(data[data["set_nan"]])

        # 保存被设置为 NaN 的行到 CSV 文件
        data[data["set_nan"]].to_csv(
            os.path.join(output_folder, "库容异常的数据行.csv"), index=False
        )
        # 绘制图形
        # plt.figure(figsize=(14, 7))
        # plt.plot(data["TM"], data["W"], label="Water Level")
        # plt.xlabel("Time")
        # plt.ylabel("Water Level (W)")
        # plt.title("Water Level Analysis with Outliers Removed")
        # plt.legend()
        # plt.show()

        cleaned_path = os.path.join(output_folder, "去除库容异常的数据.csv")
        data.to_csv(cleaned_path)
        return cleaned_path

    def back_calculation(self,data_path, file, output_folder):
        # 反推数据
        data = pd.read_csv(data_path)
        data["TM"] = pd.to_datetime(data["TM"])
        data["Time_Diff"] = data["TM"].diff().dt.total_seconds().fillna(0)
        data["INQ_ACC"] = data["OTQ"] + (10**6 * (data["W"].diff() / data["Time_Diff"]))
        data["INQ_CB"] = data["INQ"].fillna(data["INQ_ACC"])
        data["Month"] = data["TM"].dt.month
        print(data)
        data["INQ"] = data["INQ_CB"]

        back_calc_path = os.path.join(output_folder, file[:-4] + "_径流直接反推数据.csv")
        data[
            [
                "STCD",
                "TM",
                "RZ",
                "INQ",
                "W",
                "OTQ",
                "RWCHRCD",
                "RWPTN",
                "INQDR",
                "MSQMT",
                "BLRZ",
            ]
        ].to_csv(back_calc_path)
        return back_calc_path

    def delete_nan_inq(self,data_path, file, output_folder):
        # 读取CSV文件到DataFrame
        df = pd.read_csv(data_path)
        # 将'TM'列转换为日期时间格式并设置为索引
        df["TM"] = pd.to_datetime(df["TM"])

        # 设置调整后的时间为索引
        df = df.set_index("TM")

        print(df["INQ"].sum())
        # 确保'INQ'列是数值类型
        df["INQ"] = pd.to_numeric(df["INQ"], errors="coerce")

        def adjust_window(window):
            if window.count() == 0:
                return window  # 如果窗口内全是NaN，返回原窗口

            # 移除负值
            positive_values = window[window > 0]
            negative_values = window[window < 0]

            # 计算正负值的总和
            pos_sum = positive_values.sum()
            neg_sum = abs(negative_values.sum())  # 负值的绝对值和

            # 计算需要调整的比例
            if pos_sum > 0:
                adjust_factor = neg_sum / pos_sum
                # 调整正值
                adjusted_values = positive_values - (positive_values * adjust_factor)
            else:
                adjusted_values = positive_values  # 如果没有正值可用于调整，保持原样

            # 更新窗口的值
            window[window > 0] = adjusted_values
            window[window <= 0] = 0

            return window

        def rolling_with_stride(df, column, window_size, stride, func):
            # 遍历数据，步长为stride
            for i in range(0, len(df) - window_size + 1, stride):
                window_indices = range(i, i + window_size)
                df.loc[df.index[window_indices], column] = func(
                    df.loc[df.index[window_indices], column]
                )

        # 应用滚动窗口函数，这里设置步幅为4，窗口大小为7
        rolling_with_stride(df, "INQ", window_size=7, stride=4, func=adjust_window)
        path = os.path.join(output_folder, file[:-4] + "_水量平衡后的日尺度反推数据.csv")

        df["TM"] = df.index.strftime("%Y-%m-%d %H:%M:%S")
        df[
            [
                "STCD",
                "TM",
                "RZ",
                "INQ",
                "W",
                "OTQ",
                "RWCHRCD",
                "RWPTN",
                "INQDR",
                "MSQMT",
                "BLRZ",
            ]
        ].to_csv(path, index=False)
        return path

    def insert_inq(self,data_path, file, output_folder):
        # 读取CSV文件到DataFrame
        df = pd.read_csv(data_path)
        # 将'TM'列转换为日期时间格式并设置为索引
        df["TM"] = pd.to_datetime(df["TM"])
        # 设置调整后的时间为索引
        df = df.set_index("TM")
        # 确保'INQ'列是数值类型
        df["INQ"] = pd.to_numeric(df["INQ"], errors="coerce")

        # 生成从开始日期到结束日期的完整时间序列，按小时
        date_range = pd.date_range(start=df.index.min(), end=df.index.max(), freq="h")
        complete_df = pd.DataFrame(index=date_range)

        # 将原始数据与完整时间序列表格全连接
        df = complete_df.join(df, how="outer")

        # 使用线性插值
        # 插值前检查连续缺失是否超过7天（7*24小时）
        def linear_interpolate(df, column="INQ", threshold=168):
            data = df[column]
            start_index = None

            for i in range(len(data)):
                if not pd.isna(data.iloc[i]):
                    if start_index is None:
                        start_index = i
                    else:
                        # 检查当前点和上一个有数据点之间的间隔
                        if i - start_index - 1 < threshold:
                            # 如果间隔小于阈值，进行插值
                            data.iloc[start_index : i + 1] = data.iloc[
                                start_index : i + 1
                            ].interpolate()
                        # 更新起始点为当前点
                        start_index = i

            df[column] = data
            return df

        df = linear_interpolate(df)

        # 确保INQ值不小于0
        df["INQ"] = df["INQ"].clip(lower=0)

        result_path = os.path.join(output_folder, file)

        print("水量平衡的小时尺度滑动平均反推数据：输出行名称")
        print(df.columns)
        df["TM"] = df.index.strftime("%Y-%m-%d %H:%M:%S")
        df["STCD"] = df["STCD"].dropna().iloc[0]
        # 最后一步转换为整数再转换为字符串
        df["STCD"] = df["STCD"].astype(int).astype(str)
        print(df["STCD"])
        df[
            [
                "STCD",
                "TM",
                "RZ",
                "INQ",
                "W",
                "OTQ",
                "RWCHRCD",
                "RWPTN",
                "INQDR",
                "MSQMT",
                "BLRZ",
            ]
        ].to_csv(result_path, index=False)
        df[
            [
                "STCD",
                "TM",
                "RZ",
                "INQ",
                "W",
                "OTQ",
                "RWCHRCD",
                "RWPTN",
                "INQDR",
                "MSQMT",
                "BLRZ",
            ]
        ].to_csv(
            os.path.join(
                "/home/liutianxv1/水库流量数据小时插值并保持水量平衡版本", file
            ),
            index=False,
        )

        return result_path

    def process_backtrack(self):
        for file in tqdm(os.listdir(self.data_folder)):
            if file.endswith(".csv"):
                file_path = os.path.join(self.data_folder, file)
                output_folder = os.path.join(self.output_folder, file[:-4])
                if not os.path.exists(output_folder):
                    os.makedirs(output_folder)
                # Process each file step by step
                # 去除库容异常
                cleaned_data = self.clean_W(file_path, output_folder)
                # 公式计算反推
                back_data = self.back_calculation(cleaned_data, file, output_folder)
                # 去除反推异常值
                nonan_data = self.delete_nan_inq(back_data, file, output_folder)
                # 插值平衡
                #insert_data = self.insert_inq(nonan_data, file, output_folder)
                # 绘图
                



"""
Author: liutiaxqabs 1498093445@qq.com
Date: 2024-04-19 14:00:27
LastEditors: liutiaxqabs 1498093445@qq.com
LastEditTime: 2024-04-24 13:48:19
FilePath: /hydrodatasource/hydrodatasource/cleaner/waterlevel_cleaner.py
Description: 这是默认设置,请设置`customMade`, 打开koroFileHeader查看配置 进行设置: https://github.com/OBKoro1/koro1FileHeader/wiki/%E9%85%8D%E7%BD%AE
"""

import numpy as np
import pandas as pd
from .cleaner import Cleaner


class WaterlevelCleaner(Cleaner):

    def __init__(self, data_path, grad_max=200, window_size=24, *args, **kwargs):
        self.temporal_list = pd.DataFrame()  # 初始化为空的DataFrame
        self.spatial_list = pd.DataFrame()
        self.grad_max = grad_max
        self.window_size = window_size
        super().__init__(data_path, *args, **kwargs)

    def rolling_fill(self, waterlevel_data):
        df = waterlevel_data.copy()
        # 计算滑动众数，如果众数不存在使用前一个有效值
        rolling_stat = (
            df["Z"]
            .rolling(window=self.window_size, min_periods=1)
            .apply(lambda x: x.iloc[-1] if x.mode().empty else x.mode()[0], raw=False)
        )
        # 使用计算得到的滑动统计值填补缺失值
        df["Z"] = df["Z"].fillna(rolling_stat)
        return df

    def moving_gradient_filter(self, waterlevel_data):
        # 创建数据副本以避免修改原始DataFrame
        df = waterlevel_data.copy()
        # 确保时间列 "TM" 是 datetime 类型
        df["TM"] = pd.to_datetime(df["TM"])

        # 计算水位变化梯度
        df["waterlevel_Change"] = df["Z"].diff()

        # 使用滑动窗口计算每个点的局部众数，众数不存在时使用中位数
        rolling_modes = (
            df["Z"]
            .rolling(window=self.window_size, min_periods=1)
            .apply(
                lambda x: x.median() if x.mode().empty else x.mode().iloc[0],
                raw=False,
            )
        )

        # 识别汛期
        df["Is_Flood_Season"] = df["TM"].apply(lambda x: 6 <= x.month <= 9)

        # 汛期与非汛期梯度阈值
        gradient_threshold_flood = self.grad_max
        gradient_threshold_non_flood = self.grad_max / 2

        # 替换汛期中梯度过高的数据点
        flood_mask = (df["Is_Flood_Season"] == True) & (
            df["waterlevel_Change"].abs() > gradient_threshold_flood
        )
        df.loc[flood_mask, "Z"] = None  # rolling_modes[flood_mask]

        # 替换非汛期中梯度过高的数据点
        non_flood_mask = (df["Is_Flood_Season"] == False) & (
            df["waterlevel_Change"].abs() > gradient_threshold_non_flood
        )
        df.loc[non_flood_mask, "Z"] = None  # rolling_modes[non_flood_mask]

        # 清理不需要的临时列
        df.drop(columns=["waterlevel_Change", "Is_Flood_Season"], inplace=True)

        return df

    def anomaly_process(self, methods=None):
        super().anomaly_process(methods)
        waterlevel_data = self.origin_df
        for method in methods:
            if method == "moving_grad":
                waterlevel_data = self.moving_gradient_filter(
                    waterlevel_data=waterlevel_data
                )
            if method == "roll":
                waterlevel_data = self.rolling_fill(
                    waterlevel_data=waterlevel_data
                )
            else:
                print("please check your method name")

        # self.processed_df["Z"] = waterlevel_data["Z"] # 最终结果赋值给processed_df
        # 新增一列进行存储
        self.processed_df[str(methods)] = waterlevel_data["Z"]



import time

import fsspec
import geopandas as gpd
import xarray as xr

import hydrodatasource.configs.config as hdscc
import hydrodatasource.processor.mask as hpm
from hydrodatasource.reader.spliter_grid import generate_bbox_from_shp, query_path_from_metadata, \
    concat_gpm_smap_mean_data


def test_grid_mean_mask():
    # 21401550, 碧流河
    test_shp = 's3://basins-origin/basin_shapefiles/basin_CHN_songliao_21401550.zip'
    bbox, basin = generate_bbox_from_shp(test_shp, 'gpm')
    time_start = "2023-06-06 00:00:00"
    time_end = "2023-06-06 02:00:00"
    test_gpm_paths = query_path_from_metadata(time_start, time_end, bbox, data_source="gpm")
    result_arr_list = []
    for path in test_gpm_paths:
        test_gpm = xr.open_dataset(hdscc.FS.open(path))
        mask = hpm.gen_single_mask('basin_CHN_songliao_21401550', basin, 'gpm')
        result_arr = hpm.mean_by_mask(test_gpm, var='precipitationCal', mask=mask)
        result_arr_list.append(result_arr)
    return result_arr_list


def test_grid_mean_era5_land():
    # 21401550, 碧流河
    test_shp = 's3://basins-origin/basin_shapefiles/basin_CHN_songliao_21401550.zip'
    bbox, basin = generate_bbox_from_shp(test_shp, 'era5_land')
    time_start = "2022-06-02"
    time_end = "2022-06-02"
    test_era5_land_paths = query_path_from_metadata(time_start, time_end, bbox, data_source="era5_land")
    result_arr_list = []
    for path in test_era5_land_paths:
        test_era5_land = xr.open_dataset(hdscc.FS.open(path))
        mask = hpm.gen_single_mask('basin_CHN_songliao_21401550', basin, 'era5_land')
        result_arr = hpm.mean_by_mask(test_era5_land, var='tp', mask=mask)
        result_arr_list.append(result_arr)
    return result_arr_list


def test_smap_mean():
    test_shp = 's3://basins-origin/basin_shapefiles/basin_CHN_songliao_21401550.zip'
    bbox, basin = generate_bbox_from_shp(test_shp, 'smap')
    time_start = "2016-02-02"
    time_end = "2016-02-02"
    test_smap_paths = query_path_from_metadata(time_start, time_end, bbox, data_source='smap')
    result_arr_list = []
    for path in test_smap_paths:
        test_smap = xr.open_dataset(hdscc.FS.open(path))
        mask = hpm.gen_single_mask('basin_CHN_songliao_21401550', basin, 'smap')
        result_arr = hpm.mean_by_mask(test_smap, 'sm_surface', mask)
        result_arr_list.append(result_arr)
    return result_arr_list


'''
def test_concat_era5_land_average():
    basin_id = 'CHN_21401550'
    result_arr_list = test_grid_mean_era5_land()
    xr_ds = xr.Dataset(coords={'tp_aver': []})
    # xr_ds是24小时的24个雨量平均值
    for i in np.arange(0, len(result_arr_list)):
        temp_ds = xr.Dataset({'tp_aver': result_arr_list[i]})
        xr_ds = xr.concat([xr_ds, temp_ds], 'tp_aver')
    tile_path = f's3://basins-origin/hour_data/1h/grid_data/grid_era5_land_data/grid_era5_land_{basin_id}.nc'
    hdscc.FS.write_bytes(tile_path, xr_ds.to_netcdf())
    return xr_ds
'''


def test_concat_variables():
    # '3B-HHR-E.MS.MRG.3IMERG.20200701-S000000-E002959.0000.V06B.HDF5_tile.nc4'未区分流域，导致前面的数据被后面的数据覆盖
    basin_ids = ['basin_CHN_songliao_21401550']
    merge_list = concat_gpm_smap_mean_data(basin_ids, [['2022-06-01 00:00:00', '2022-07-31 23:00:00'],
                                                       ['2022-09-01 00:00:00', '2022-09-30 23:00:00']])
    for xr_ds in merge_list:
        hdscc.FS.write_bytes(f's3://basins-origin/hour_data/1h/mean_data/mean_data_merged/mean_data_{basin_ids[0]}',
                             xr_ds.to_netcdf())


def test_concat_basins_variables():
    # 21100150、21110150、21110400、21113800
    # (stations-origin/zz_stations/hour_data/1h/zz_CHN_songliao_{stcd}.csv)数据只有2021.4-2021.7，基本是废站
    basin_ids = ['basin_CHN_songliao_10810201', 'basin_CHN_songliao_20501500', 'basin_CHN_songliao_21113800']
    merge_list = concat_gpm_smap_mean_data(basin_ids, [['2020-07-01 00:00:00', '2020-07-31 23:00:00']])
    for i in range(0, merge_list):
        hdscc.FS.write_bytes(f's3://basins-origin/hour_data/1h/mean_data/mean_data_{basin_ids[i]}',
                             merge_list[i].to_netcdf())


def test_concat_usa_basins_variables():
    basin_usa_ids = ['basin_USA_camels_01491000', 'basin_USA_camels_01548500', 'basin_USA_camels_02014000',
                     'basin_USA_camels_02046000', 'basin_USA_camels_02051500']
    merge_list = concat_gpm_smap_mean_data(basin_usa_ids,
                                           [['2022-07-01 00:00:00', '2022-09-30 23:00:00'],
                                            ['2023-07-01 00:00:00', '2023-09-30 23:00:00']])
    '''
    for i in range(0, merge_list):
        hdscc.FS.write_bytes(f's3://basins-origin/hour_data/1h/mean_data/mean_data_{basin_usa_ids[i]}',
                             merge_list[i].to_netcdf())
    '''
    return merge_list


def test_read_era5_by_basins():
    data_name = 'era5'
    chk_dir = 'single_levels_tp'
    # era5中的经度是0-360°，对西经如美国会出问题
    basin_ids = ['21401550']
    test_chk_dir = f's3://era5-origin/era5/grib/{chk_dir}/2016/02/02/'
    test_paths = hdscc.FS.glob(test_chk_dir + '**')[1:]
    basin_tgdf = gpd.read_file(hdscc.FS.open('s3://basins-origin/basins_shp.zip'))
    basin_gdfs = basin_tgdf[basin_tgdf['BASIN_ID'].isin(basin_ids)]
    basin_mean_dict = {}
    time_mean_list = []
    start_cyc_time = time.time()
    for basin_id in basin_ids:
        for time_path in test_paths:
            basin_part = basin_gdfs[basin_gdfs['BASIN_ID'] == basin_id]
            mask = hpm.gen_single_mask(basin_id, basin_part, data_name)
            bbox = [mask['lon'].values.min(), mask['lon'].values.max(), mask['lat'].values.max(),
                    mask['lat'].values.min()]
            time_path = f'simplecache::s3://{time_path}'
            time_ds = xr.open_dataset(fsspec.open_local(time_path, s3=hdscc.MINIO_PARAM, filecache=
                {'cache_storage': '/tmp/files'}), engine='cfgrib')
            time_ds_split = time_ds.sel(longitude=slice(bbox[0], bbox[1]), latitude=slice(bbox[2], bbox[3]))
            mean_res = hpm.mean_by_mask(src=time_ds_split, var='tp', mask=mask)
            time_mean_list.append(mean_res)
        basin_mean_dict[basin_id] = time_mean_list
    use_time = time.time() - start_cyc_time
    print(use_time)
    return basin_mean_dict

'''
def split_dataset(time_path, bboxes: list):
    time_path = f'simplecache::s3://{time_path}'
    time_ds = xr.open_dataset(fsspec.open_local(time_path, s3=hdscc.MINIO_PARAM, filecache=
    {'cache_storage': '/tmp/files'}), engine='cfgrib')
    datasets = [time_ds.sel(longitude=slice(bbox[0], bbox[1]), latitude=slice(bbox[2], bbox[3])) for bbox in bboxes]
    result = xr.concat(datasets, dim='time')
    return result
'''



import os
import pathlib

import geopandas as gpd
import pandas as pd
import xarray as xr
from hydrotopo import ig_path
import hydrodatasource.configs.config as conf
from hydrodatasource.reader.spliter_grid import (
    query_path_from_metadata,
    generate_bbox_from_shp,
)


def test_generate_bbox_from_shp():
    basin_shp = "s3://basins-origin/basins_shp.zip"
    bbox = generate_bbox_from_shp(basin_shape_path=basin_shp, data_source="gpm")
    return bbox


def test_split_grid_data_from_single_basin_gpm():
    test_shp = "s3://basins-origin/basin_shapefiles/basin_USA_camels_12145500.zip"
    bbox = generate_bbox_from_shp(test_shp, data_source="gpm")
    time_start = "2018-06-05 01:00:00"
    time_end = "2018-06-05 02:00:00"
    tile_list = query_path_from_metadata(time_start, time_end, bbox, data_source="gpm")
    data_list = []
    for tile in tile_list:
        data_list.append(xr.open_dataset(conf.FS.open(tile)))
    print(data_list)
    return tile_list


def test_split_grid_data_from_single_basin_gfs():
    test_shp = "s3://basins-origin/basin_shapefiles/basin_USA_camels_01414500.zip"
    bbox = generate_bbox_from_shp(test_shp, data_source="gfs")
    time_start = "2022-01-03"
    time_end = "2022-01-03"
    tile_list = query_path_from_metadata(time_start, time_end, bbox, data_source="gfs")
    data_list = []
    for tile in tile_list:
        data_list.append(xr.open_dataset(conf.FS.open(tile)))
    print(data_list)
    return tile_list


def test_split_grid_data_from_single_basin_smap():
    test_shp = "s3://basins-origin/basin_shapefiles/basin_USA_camels_01414500.zip"
    bbox = generate_bbox_from_shp(test_shp, data_source="smap")
    time_start = "2016-02-02"
    time_end = "2016-02-02"
    tile_list = query_path_from_metadata(time_start, time_end, bbox, data_source="smap")
    data_list = []
    for tile in tile_list:
        data_list.append(xr.open_dataset(conf.FS.open(tile)))
    print(data_list)
    return tile_list


def test_split_grid_data_from_single_basin_era5():
    test_shp = "s3://basins-origin/basin_shapefiles/basin_CHN_songliao_10810201.zip"
    bbox = generate_bbox_from_shp(test_shp, data_source="era5_land")
    time_start = "2022-06-02"
    time_end = "2022-06-02"
    tile_list = query_path_from_metadata(
        time_start, time_end, bbox, data_source="era5_land"
    )
    data_list = []
    for tile in tile_list:
        data_list.append(xr.open_dataset(conf.FS.open(tile)))
    print(data_list)
    return tile_list


def test_read_topo_data():
    dams_shp = gpd.read_file(
        conf.FS.open("s3://reservoirs-origin/dams.zip"), engine="pyogrio"
    )
    network_shp = gpd.read_file(
        os.path.join(
            pathlib.Path(__file__).parent.parent,
            "data/river_network/songliao_cut_single.shp",
        ),
        engine="pyogrio",
    )
    index = dams_shp.index[dams_shp["ID"] == "zq_CHN_songliao_10310500"]
    paths = ig_path.find_edge_nodes(dams_shp, network_shp, index, "up")
    for station in paths:
        sta_id = dams_shp["ID"][dams_shp.index == station].to_list()[0]
        rr_path = "s3://reservoirs-origin/rr_stations/" + sta_id + ".csv"
        rr_df = pd.read_csv(rr_path, storage_options=conf.MINIO_PARAM)
        print(rr_df)



"""
Author: Yang Wang
Date: 2024-03-16 15:55:22
LastEditTime: 2024-03-28 08:40:55
LastEditors: Wenyu Ouyang
Description: 
FilePath: \hydrodatasource\tests\test_read_blob.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""
import time
import intake
import ujson
import xarray as xr
import zarr
from distributed import LocalCluster
from kerchunk.hdf import SingleHdf5ToZarr

import hydrodatasource.configs.config as conf


# @pytest.mark.asyncio
def test_read_blob():
    """
    start_time = time.time()
    cluster = LocalCluster(dashboard_address=":10086")  # Fully-featured local Dask cluster
    client = cluster.get_client()
    # ds = xr.open_dataset('/home/forestbat/usgs-streamflow-nldas_hourly.nc')
    ds = xr.open_dataset('/home/forestbat/gpm_gfs.nc')
    # new_zarr = ds.to_zarr('usgs-streamflow-nldas_hourly.zarr', mode='w')
    print(time.time() - start_time)
    """
    """
    zarr_start_time = time.time()
    # https://pastebin.com/fKKECf3U
    zarr_path = conf.FS.get_mapper('s3://datasets-origin/usgs_streamflow_nldas_hourly.zarr')
    wrapped_store = zarr.storage.KVStore(zarr_path)
    zds = xr.open_zarr(wrapped_store)
    print(time.time() - zarr_start_time)
    ###########################
    cluster = LocalCluster(dashboard_address=":10086")  # Fully-featured local Dask cluster
    client = cluster.get_client()
    dask_zarr_start_time = time.time()
    zarr_dask_path = conf.FS.get_mapper('s3://datasets-origin/usgs_streamflow_nldas_hourly.zarr')
    zarr_dask_store = zarr.storage.KVStore(zarr_dask_path)
    dzds = xr.open_zarr(zarr_dask_store)
    print(time.time() - dask_zarr_start_time)
    """
    ###########################
    """
    # 5~7 min
    nc_to_zarr_time = time.time()
    nc_chunks = SingleHdf5ToZarr(h5f='/home/forestbat/usgs-streamflow-nldas_hourly.nc')
    with open('gpm_gfs_nc.json', 'wb') as fpj:
        fpj.write(ujson.dumps(nc_chunks.translate()).encode())
    print(time.time() - nc_to_zarr_time)
    """
    ###########################
    virtual_zarr_start_time = time.time()
    """
    # Method 0
    vds = xr.open_dataset("reference://", engine="zarr", backend_kwargs={
        "storage_options": {"fo": 'gpm_gfs_nc.json'},
        "remote_options": conf.MINIO_PARAM,
        "remote_protocol": "s3",
    })
    """
    # ~2mints
    vds = xr.open_dataset("gpm_gfs_nc.json", engine="kerchunk")
    print(time.time() - virtual_zarr_start_time)



'''
Author: liutiaxqabs 1498093445@qq.com
Date: 2024-04-19 13:58:31
LastEditors: liutiaxqabs 1498093445@qq.com
LastEditTime: 2024-04-26 17:09:33
FilePath: /hydrodatasource/hydrodatasource/cleaner/cleaner.py
Description: 这是默认设置,请设置`customMade`, 打开koroFileHeader查看配置 进行设置: https://github.com/OBKoro1/koro1FileHeader/wiki/%E9%85%8D%E7%BD%AE
'''
"""
cleaner/
│
├── __init__.py
├── cleaner.py          # 包含 Cleaner 基类
├── rainfall_cleaner.py # 包含 RainfallCleaner 类
├── streamflow_cleaner.py # 包含 StreamflowCleaner 类
└── waterlevel_cleaner.py # 包含 WaterlevelCleaner 类
"""

import xarray as xr
import pandas as pd
import numpy as np


class Cleaner:
    def __init__(self, data_path, *args, **kwargs):
        self.data_path = data_path
        self.origin_df = None
        self.processed_df = None
        self.read_data()

    def read_data(self):
        # 读取数据并存储在origin_df中
        self.origin_df = pd.read_csv(self.data_path, dtype={"STCD": str}, index_col=False)
        self.processed_df = self.origin_df.copy()

    def save_data(self, data, output_path):
        # 保存数据到CSV
        data.to_csv(output_path, index=False)

    def anomaly_process(self, methods=None):
        if methods is None:
            methods = []
        # 如果有特定流程，可以在这里添加
        pass
